import pathlib
from typing import Any

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.base import Project as OpenSvipProject

from .options import InputOptions, OutputOptions


class JsonSvipConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        return OpenSvipProject.model_validate_json(path.read_bytes().decode("utf-8"))

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        dump_kwargs: dict[str, Any] = (
            {"indent": 2} if options.indented else {"separators": (",", ":")}
        )
        path.write_bytes(
            json.dumps(
                project.model_dump(mode="json", by_alias=True),
                ensure_ascii=False,
                **dump_kwargs,
            ).encode("utf-8"),
        )



from pydantic import Field

from libresvip.model.base import BaseModel

from .models.vsqx3 import MasterTrack as Vsq3MasterTrack
from .models.vsqx3 import MCtrl as Vsq3MCtrl
from .models.vsqx3 import Mixer as Vsq3Mixer
from .models.vsqx3 import MonoTrack as Vsq3MonoTrack
from .models.vsqx3 import MonoUnit as Vsq3MonoUnit
from .models.vsqx3 import MusicalPart as Vsq3MusicalPart
from .models.vsqx3 import Note as Vsq3Note
from .models.vsqx3 import NoteStyle as Vsq3NoteStyle
from .models.vsqx3 import ParameterNames as Vsq3ParameterNames
from .models.vsqx3 import PartStyle as Vsq3PartStyle
from .models.vsqx3 import Singer as Vsq3Singer
from .models.vsqx3 import StereoTrack as Vsq3StereoTrack
from .models.vsqx3 import StereoUnit as Vsq3StereoUnit
from .models.vsqx3 import Tempo as Vsq3Tempo
from .models.vsqx3 import TimeSig as Vsq3TimeSig
from .models.vsqx3 import TypeParamAttr as Vsq3TypeParamAttr
from .models.vsqx3 import TypePhonemes as Vsq3TypePhonemes
from .models.vsqx3 import Vsq3
from .models.vsqx3 import VsTrack as Vsq3VsTrack
from .models.vsqx3 import VsUnit as Vsq3VsUnit
from .models.vsqx3 import VVoice as Vsq3VVoice
from .models.vsqx3 import VVoiceTable as Vsq3VVoiceTable
from .models.vsqx3 import WavPart as Vsq3WavPart
from .models.vsqx4 import MasterTrack as Vsq4MasterTrack
from .models.vsqx4 import MCtrl as Vsq4MCtrl
from .models.vsqx4 import Mixer as Vsq4Mixer
from .models.vsqx4 import MonoTrack as Vsq4MonoTrack
from .models.vsqx4 import MonoUnit as Vsq4MonoUnit
from .models.vsqx4 import MusicalPart as Vsq4MusicalPart
from .models.vsqx4 import Note as Vsq4Note
from .models.vsqx4 import NoteStyle as Vsq4NoteStyle
from .models.vsqx4 import ParameterNames as Vsq4ParameterNames
from .models.vsqx4 import PartStyle as Vsq4PartStyle
from .models.vsqx4 import Singer as Vsq4Singer
from .models.vsqx4 import StereoTrack as Vsq4StereoTrack
from .models.vsqx4 import StereoUnit as Vsq4StereoUnit
from .models.vsqx4 import Tempo as Vsq4Tempo
from .models.vsqx4 import TimeSig as Vsq4TimeSig
from .models.vsqx4 import TypeParamAttr as Vsq4TypeParamAttr
from .models.vsqx4 import TypePhonemes as Vsq4TypePhonemes
from .models.vsqx4 import Vsq4
from .models.vsqx4 import VsTrack as Vsq4VsTrack
from .models.vsqx4 import VsUnit as Vsq4VsUnit
from .models.vsqx4 import VVoice as Vsq4VVoice
from .models.vsqx4 import VVoiceTable as Vsq4VVoiceTable
from .models.vsqx4 import WavPart as Vsq4WavPart

Vsqx = Vsq3 | Vsq4
VsqxMasterTrack = Vsq3MasterTrack | Vsq4MasterTrack
VsqxMCtrl = Vsq3MCtrl | Vsq4MCtrl
VsqxMixer = Vsq3Mixer | Vsq4Mixer
VsqxMonoTrack = Vsq3MonoTrack | Vsq4MonoTrack
VsqxMusicalPart = Vsq3MusicalPart | Vsq4MusicalPart
VsqxNote = Vsq3Note | Vsq4Note
VsqxNoteStyle = Vsq3NoteStyle | Vsq4NoteStyle
VsqxParameterNames = Vsq3ParameterNames | Vsq4ParameterNames
VsqxPartStyle = Vsq3PartStyle | Vsq4PartStyle
VsqxTypePhonemes = Vsq3TypePhonemes | Vsq4TypePhonemes
VsqxSinger = Vsq3Singer | Vsq4Singer
VsqxStereoTrack = Vsq3StereoTrack | Vsq4StereoTrack
VsqxStereoUnitList = list[Vsq3StereoUnit] | list[Vsq4StereoUnit]
VsqxMonoUnitList = list[Vsq3MonoUnit] | list[Vsq4MonoUnit]
VsqxWavUnitList = VsqxStereoUnitList | VsqxMonoUnitList
VsqxTempoList = list[Vsq3Tempo] | list[Vsq4Tempo]
VsqxTimeSigList = list[Vsq3TimeSig] | list[Vsq4TimeSig]
VsqxTypeParamAttr = Vsq3TypeParamAttr | Vsq4TypeParamAttr
VsqxVsTrackList = list[Vsq3VsTrack] | list[Vsq4VsTrack]
VsqxVsUnitList = list[Vsq3VsUnit] | list[Vsq4VsUnit]
VsqxVVoice = Vsq3VVoice | Vsq4VVoice
VsqxVVoiceTable = Vsq3VVoiceTable | Vsq4VVoiceTable
VsqxWavPartList = list[Vsq3WavPart] | list[Vsq4WavPart]


class VocaloidStyleTypes(BaseModel):
    accent: int = 50
    bend_depth: int = Field(0, alias="bendDep")
    bend_length: int = Field(0, alias="bendLen")
    decay: int = 50
    fall_portamento: int = Field(0, alias="fallPort")
    opening: int = 127
    rise_portamento: int = Field(0, alias="risePort")
    vibrato_length: int = Field(0, alias="vibLen")
    vibrato_type: int = Field(0, alias="vibType")
    vibrato_depth: list[int] | None = Field(None, alias="vibDep")
    vibrato_rate: list[int] | None = Field(None, alias="vibRate")


__all__ = [
    "VocaloidStyleTypes",
    "Vsq3",
    "Vsq3MCtrl",
    "Vsq3MasterTrack",
    "Vsq3Mixer",
    "Vsq3MonoTrack",
    "Vsq3MonoUnit",
    "Vsq3MusicalPart",
    "Vsq3Note",
    "Vsq3NoteStyle",
    "Vsq3ParameterNames",
    "Vsq3PartStyle",
    "Vsq3Singer",
    "Vsq3StereoTrack",
    "Vsq3StereoUnit",
    "Vsq3Tempo",
    "Vsq3TimeSig",
    "Vsq3TypeParamAttr",
    "Vsq3TypePhonemes",
    "Vsq3VVoice",
    "Vsq3VVoiceTable",
    "Vsq3VsTrack",
    "Vsq3VsUnit",
    "Vsq3WavPart",
    "Vsq4",
    "Vsq4MCtrl",
    "Vsq4MasterTrack",
    "Vsq4Mixer",
    "Vsq4MonoTrack",
    "Vsq4MonoUnit",
    "Vsq4MusicalPart",
    "Vsq4Note",
    "Vsq4NoteStyle",
    "Vsq4ParameterNames",
    "Vsq4PartStyle",
    "Vsq4Singer",
    "Vsq4StereoTrack",
    "Vsq4StereoUnit",
    "Vsq4Tempo",
    "Vsq4TimeSig",
    "Vsq4TypeParamAttr",
    "Vsq4TypePhonemes",
    "Vsq4VVoice",
    "Vsq4VVoiceTable",
    "Vsq4VsTrack",
    "Vsq4VsUnit",
    "Vsq4WavPart",
    "Vsqx",
    "VsqxMCtrl",
    "VsqxMasterTrack",
    "VsqxMixer",
    "VsqxMonoTrack",
    "VsqxMonoUnitList",
    "VsqxMusicalPart",
    "VsqxNote",
    "VsqxNoteStyle",
    "VsqxParameterNames",
    "VsqxPartStyle",
    "VsqxSinger",
    "VsqxStereoTrack",
    "VsqxStereoUnitList",
    "VsqxTempoList",
    "VsqxTimeSigList",
    "VsqxTypeParamAttr",
    "VsqxTypePhonemes",
    "VsqxVVoice",
    "VsqxVVoiceTable",
    "VsqxVsTrackList",
    "VsqxVsUnitList",
    "VsqxWavPartList",
    "VsqxWavUnitList",
]



import fractions

import more_itertools

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project, SingingTrack

from .options import ProcessOptions


class RemoveShortSilencesMiddleware(plugin_base.MiddlewareBase):
    def process(self, project: Project, options: ProcessOptions) -> Project:
        first_bar_tick = round(project.time_signature_list[0].bar_length())
        if min_silence_length := fractions.Fraction(options.fill_threshold.value):
            silence_threshold = first_bar_tick * min_silence_length
            for track in project.track_list:
                if isinstance(track, SingingTrack):
                    for prev_note, note in more_itertools.pairwise(track.note_list):
                        if 0 < note.start_pos - prev_note.end_pos < silence_threshold:
                            prev_note.length = note.start_pos - prev_note.start_pos
        return project



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .lrc_generator import LrcGenerator
from .options import OutputOptions
from .template import render_lrc


class LrcConverter(plugin_base.WriteOnlyConverterBase):
    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        lrc_model = LrcGenerator(options).generate_project(project)
        render_lrc(lrc_model, path)



from importlib.resources import as_file

from PySide6 import QtCore

from libresvip.core.constants import res_dir


def init_resources() -> None:
    with as_file(res_dir / "resources.rcc") as rcc_file:
        QtCore.QResource.registerResource(str(rcc_file))


def cleanup_resources() -> None:
    with as_file(res_dir / "resources.rcc") as rcc_file:
        QtCore.QResource.unregisterResource(str(rcc_file))


init_resources()



from __future__ import annotations

import sys

from loguru import logger

from libresvip.web.pages import main


def unraisablehook(unraisable: sys.UnraisableHookArgs) -> None:
    if unraisable.exc_type is not KeyboardInterrupt:
        logger.exception(unraisable.exc_value)
    del unraisable


sys.unraisablehook = unraisablehook

if __name__ in {"__main__", "__mp_main__"}:
    main()



from pydantic import BaseModel

from libresvip.model.option_mixins import EnablePitchImportationMixin


class InputOptions(EnablePitchImportationMixin, BaseModel):
    pass


class OutputOptions(BaseModel):
    pass



import dataclasses

from libresvip.core.tick_counter import skip_tempo_list
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)

from .model import (
    UFData,
    UFNotes,
    UFPitch,
    UFProject,
    UFTempos,
    UFTimeSignatures,
    UFTracks,
)
from .options import OutputOptions


@dataclasses.dataclass
class UFDataGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> UFData:
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        return UFData(
            project=UFProject(
                tempos=self.generate_tempos(project.song_tempo_list),
                time_signatures=self.generate_time_signatures(project.time_signature_list),
                tracks=self.generate_tracks(project.track_list),
                measure_prefix=0,
            )
        )

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> list[UFTempos]:
        song_tempo_list = skip_tempo_list(song_tempo_list, self.first_bar_length)
        return [
            UFTempos(
                tick_position=tempo.position,
                bpm=tempo.bpm,
            )
            for tempo in song_tempo_list
        ]

    @staticmethod
    def generate_time_signatures(
        time_signature_list: list[TimeSignature],
    ) -> list[UFTimeSignatures]:
        return [
            UFTimeSignatures(
                measure_position=time_signature.bar_index,
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            for time_signature in time_signature_list
        ]

    def generate_tracks(self, track_list: list[Track]) -> list[UFTracks]:
        return [
            UFTracks(
                name=track.title,
                notes=self.generate_notes(track.note_list),
                pitch=self.generate_pitch(track.edited_params.pitch, track.note_list),
            )
            for track in track_list
            if isinstance(track, SingingTrack)
        ]

    @staticmethod
    def generate_notes(note_list: list[Note]) -> list[UFNotes]:
        return [
            UFNotes(
                tick_on=note.start_pos,
                tick_off=note.end_pos,
                lyric=note.lyric,
                key=note.key_number,
            )
            for note in note_list
        ]

    def generate_pitch(self, pitch: ParamCurve, notes: list[Note]) -> UFPitch:
        uf_pitch = UFPitch(
            is_absolute=True,
            ticks=[],
            values=[],
        )
        if notes:
            for point in pitch.points.root:
                if point.x not in (-192000, 1073741823):
                    uf_pitch.ticks.append(point.x - self.first_bar_length)
                    uf_pitch.values.append(0 if point.y == -100 else point.y / 100)
        return uf_pitch



import dataclasses
from typing import cast

from libresvip.core.lyric_phoneme.chinese import CHINESE_RE, get_pinyin_series
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.text import LATIN_ALPHABET

from .model import (
    MutaAudioTrackData,
    MutaNote,
    MutaNoteTiming,
    MutaParams,
    MutaPoint,
    MutaProject,
    MutaSongTrackData,
    MutaTempo,
    MutaTimeSignature,
    MutaTrack,
    MutaTrackType,
)
from .options import OutputOptions


@dataclasses.dataclass
class MutaGenerator:
    options: OutputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> MutaProject:
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        time_signatures = self.generate_time_signatures(project.time_signature_list)
        tempos = self.generate_tempos(project.song_tempo_list)
        singing_tracks = self.generate_singing_tracks(
            [track for track in project.track_list if isinstance(track, SingingTrack)]
        )
        instrumental_tracks = self.generate_instrumental_tracks(
            [track for track in project.track_list if isinstance(track, InstrumentalTrack)],
            len(singing_tracks),
        )

        return MutaProject(
            file_version=101,
            time_signatures=time_signatures,
            tempos=tempos,
            tracks=singing_tracks + instrumental_tracks,
        )

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature]
    ) -> list[MutaTimeSignature]:
        return [
            MutaTimeSignature(
                measure_position=max(time_signature.bar_index - 1, 0),
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            for time_signature in time_signatures
        ]

    def generate_tempos(self, tempos: list[SongTempo]) -> list[MutaTempo]:
        return [
            MutaTempo(
                position=max(tempo.position - self.first_bar_length, 0),
                bpm=round(tempo.bpm * 100),
            )
            for tempo in tempos
        ]

    def generate_singing_tracks(self, tracks: list[SingingTrack]) -> list[MutaTrack]:
        track_list: list[MutaTrack] = []
        for track in tracks:
            muta_track = MutaTrack(
                track_type=cast("MutaTrackType", MutaTrackType.SONG),
                seq_count=1,
                name=f"Song{len(track_list) + 1}",
                mute=track.mute,
                solo=track.solo,
                volume=50,
                pan=50,
                padding=b"\x00" * 52,
                talk_track_data=None,
                audio_track_data=None,
                song_track_data=[
                    MutaSongTrackData(
                        start=self.first_bar_length,
                        length=max(
                            (note.end_pos for note in track.note_list),
                            default=0,
                        )
                        + self.first_bar_length,
                        singer_name=[ord(c) for c in self.options.default_singer_name]
                        + [0] * (258 - len(self.options.default_singer_name)),
                        unknown_1=0,
                        notes=self.generate_notes(track.note_list),
                        params=MutaParams(
                            unknown_param=[],
                            pitch_range=[],
                            pitch_data=[],
                            volume_data=[],
                            vibrato_amplitude_range=[],
                            vibrato_amplitude_data=[],
                            vibrato_frequency_range=[],
                            vibrato_frequency_data=[],
                        ),
                    )
                ],
            )
            if muta_track.song_track_data is not None and (
                pitch_points := self.generate_pitch(track.edited_params.pitch)
            ):
                muta_track.song_track_data[0].params.pitch_data = pitch_points
            track_list.append(muta_track)
        return track_list

    def generate_pitch(self, pitch: ParamCurve) -> list[MutaPoint]:
        return [
            MutaPoint(
                time=max(point.x - 2 * self.first_bar_length, -1),
                value=12900 if point.y < 0 else point.y - 1200,
            )
            for point in pitch.points.root
        ]

    def generate_notes(self, notes: list[Note]) -> list[MutaNote]:
        return [
            MutaNote(
                start=note.start_pos - self.first_bar_length,
                length=note.length,
                key=139 - note.key_number,
                lyric=[ord(c) for c in note.lyric] + [0] * (8 - len(note.lyric)),
                phoneme=self.generate_phoneme(note),
                tmg_data=[MutaNoteTiming(ori_pos=0, mod_pos=0)] * 5,
            )
            for note in notes
        ]

    @staticmethod
    def generate_phoneme(note: Note) -> str:
        result = ""
        if note.lyric == "-":
            result = "-"
        elif LATIN_ALPHABET.fullmatch(note.pronunciation or note.lyric) is not None:
            result = note.pronunciation or note.lyric
        elif CHINESE_RE.fullmatch(note.lyric) is not None:
            result = " ".join(get_pinyin_series(note.lyric))
        if len(result.encode("utf-16-le")) > 16:
            result = ""
        return result

    def generate_instrumental_tracks(
        self, tracks: list[InstrumentalTrack], singing_track_count: int
    ) -> list[MutaTrack]:
        track_list: list[MutaTrack] = []
        for track in tracks:
            if (track_info := audio_track_info(track.audio_file_path, only_wav=True)) is not None:
                muta_track = MutaTrack(
                    track_type=cast("MutaTrackType", MutaTrackType.AUDIO),
                    seq_count=1,
                    name=f"Audio{len(track_list) + 1}",
                    mute=track.mute,
                    solo=track.solo,
                    volume=50,
                    pan=50,
                    padding=b"\x00" * 52,
                    talk_track_data=None,
                    song_track_data=None,
                    audio_track_data=[
                        MutaAudioTrackData(
                            start=track.offset,
                            length=round(
                                self.time_synchronizer.get_actual_ticks_from_secs_offset(
                                    track.offset, track_info.duration / 1000
                                )
                            ),
                            file_path=track.audio_file_path + "\0",
                        )
                    ],
                )
                track_list.append(muta_track)
        return track_list



import pathlib

from jinja2 import Template

from .model import NNProject

NN_TEMPLATE = Template(
    """\
{{ nn_project.info_line.tempo }} {{ nn_project.info_line.time_signature.numerator }} {{ nn_project.info_line.time_signature.denominator }} {{ nn_project.info_line.bar_count }} {{ nn_project.info_line.version }} {{ nn_project.info_line.unknown }} 0 0 0 0
{{ nn_project.note_count }}
{% for note in nn_project.notes %}\
 {{ note.lyric }} {{ note.pronunciation }} {{ note.start }} {{ note.duration }} {{ note.key }} {{ note.cle }} {{ note.vel }} {{ note.por }} {{ note.vibrato_length }} {{ note.vibrato_depth }} {{ note.vibrato_rate }} {{ note.dynamics.point_count }},{{ note.dynamics.points|join(',') }} {{ note.pitch.point_count }},{{ note.pitch.points|join(',') }} {{ note.pitch_bend_sensitivity }}
{% endfor %}\
"""
)


def render_nn(nn_project: NNProject, output_path: pathlib.Path) -> None:
    output_path.write_bytes(NN_TEMPLATE.render(nn_project=nn_project).encode("utf-8"))



import dataclasses
import functools
import math
from collections.abc import Callable
from typing import NamedTuple

import more_itertools
import portion

from libresvip.core.exceptions import NotesOverlappedError, ParamsError
from libresvip.core.time_interval import PiecewiseIntervalDict
from libresvip.utils.music_math import clamp, linear_interpolation
from libresvip.utils.search import find_index
from libresvip.utils.translation import gettext_lazy as _

from .constants import MAX_BREAK
from .interval_utils import position_to_ticks
from .lambert_w import LambertW
from .model import SVCurvePitchControl, SVPitchControl


class NoteStruct(NamedTuple):
    key: int
    start: float
    end: float
    portamento_offset: float
    portamento_left: float
    portamento_right: float
    depth_left: float
    depth_right: float
    vibrato_start: float
    vibrato_left: float
    vibrato_right: float
    vibrato_depth: float
    vibrato_frequency: float
    vibrato_phase: float


@dataclasses.dataclass
class SigmoidNode:
    start: float = dataclasses.field(init=False)
    end: float = dataclasses.field(init=False)
    center: float = dataclasses.field(init=False)
    sigmoid_l: Callable[[float], float] = dataclasses.field(init=False)
    sigmoid_r: Callable[[float], float] = dataclasses.field(init=False)
    d_sigmoid_l: Callable[[float], float] = dataclasses.field(init=False)
    d_sigmoid_r: Callable[[float], float] = dataclasses.field(init=False)
    _start: dataclasses.InitVar[float]
    _end: dataclasses.InitVar[float]
    _center: dataclasses.InitVar[float]
    _radius: dataclasses.InitVar[float]
    _key_left: dataclasses.InitVar[int]
    _key_right: dataclasses.InitVar[int]

    @property
    def k(self) -> float:
        return 5.5

    def __post_init__(
        self,
        _start: float,
        _end: float,
        _center: float,
        _radius: float,
        _key_left: int,
        _key_right: int,
    ) -> None:
        self.start = _start
        self.end = _end
        self.center = _center
        h: float = (_key_right - _key_left) * 100
        a = 1 / (1 + math.exp(self.k))
        power = 0.75

        left = self.center - self.start
        if left >= _radius or left == 0:
            k_l = self.k
            h_l = h
            d_l = 0.0
        else:
            al = a * math.pow(_radius / left, power)
            bl = left / _radius
            cl = al * bl * self.k / (2 * al - 1)
            k_l = al / (2 * al - 1) * self.k - 1 / bl * LambertW.evaluate(cl * math.exp(cl), -1)
            h_l = h * k_l / (2 * k_l - self.k)
            d_l = -_radius / k_l * math.log(2 * h_l / h - 1)

        self.sigmoid_l = lambda x: _key_left * 100 + h_l / (
            1 + math.exp(-k_l / _radius * (x - self.center + d_l))
        )
        self.d_sigmoid_l = (
            lambda x: h_l
            * k_l
            / _radius
            * math.exp(-k_l / _radius * (x - self.center + d_l))
            / math.pow(1 + math.exp(-k_l / _radius * (x - self.center + d_l)), 2)
        )

        right = self.end - self.center
        if right >= _radius or right == 0:
            k_r = self.k
            h_r = h
            d_r = 0.0
        else:
            ar = a * math.pow(_radius / right, power)
            br = right / _radius
            cr = ar * br * self.k / (2 * ar - 1)
            k_r = ar / (2 * ar - 1) * self.k - 1 / br * LambertW.evaluate(cr * math.exp(cr), -1)
            h_r = h * k_r / (2 * k_r - self.k)
            d_r = -_radius / k_r * math.log(2 * h_r / h - 1)

        self.sigmoid_r = lambda x: _key_right * 100 - h_r / (
            1 + math.exp(-k_r / _radius * (self.center - x + d_r))
        )
        self.d_sigmoid_r = (
            lambda x: h_r
            * k_r
            / _radius
            * math.exp(-k_r / _radius * (self.center - x + d_r))
            / math.pow(1 + math.exp(-k_r / _radius * (self.center - x + d_r)), 2)
        )

    def value_at_secs(self, secs: float) -> float:
        return self.sigmoid_l(secs) if secs <= self.center else self.sigmoid_r(secs)

    def slope_at_secs(self, secs: float) -> float:
        return self.d_sigmoid_l(secs) if secs <= self.center else self.d_sigmoid_r(secs)


@dataclasses.dataclass
class BaseLayerGenerator:
    default_radius = 0.07
    _note_list: dataclasses.InitVar[list[NoteStruct]]
    note_list: list[NoteStruct] = dataclasses.field(init=False)
    sigmoid_nodes: list[SigmoidNode] = dataclasses.field(default_factory=list)

    def __post_init__(self, _note_list: list[NoteStruct]) -> None:
        if not _note_list:
            return
        self.note_list = _note_list
        for current_note, next_note in more_itertools.pairwise(self.note_list):
            if current_note.key == next_note.key:
                continue
            elif current_note.end > next_note.start:
                msg = _("Notes Overlapped")
                raise NotesOverlappedError(msg)
            if (diameter := current_note.portamento_right + next_note.portamento_left) and (
                next_note.start - current_note.end <= MAX_BREAK
            ):
                start = clamp(
                    current_note.end - current_note.portamento_right + next_note.portamento_offset,
                    current_note.start,
                    current_note.end,
                )
                end = clamp(
                    next_note.start + next_note.portamento_left + next_note.portamento_offset,
                    next_note.start,
                    next_note.end,
                )
                mid = clamp(
                    (current_note.end + next_note.start) / 2 + next_note.portamento_offset,
                    start,
                    end,
                )
                self.sigmoid_nodes.append(
                    SigmoidNode(
                        _start=start,
                        _end=end,
                        _center=mid,
                        _radius=diameter / 2,
                        _key_left=current_note.key,
                        _key_right=next_note.key,
                    )
                )
            else:
                mid = (current_note.end + next_note.start) / 2
                self.sigmoid_nodes.append(
                    SigmoidNode(
                        _start=mid - self.default_radius,
                        _end=mid + self.default_radius,
                        _center=mid,
                        _radius=self.default_radius,
                        _key_left=current_note.key,
                        _key_right=next_note.key,
                    )
                )

    def pitch_at_secs(self, secs: float) -> float:
        query = [node for node in self.sigmoid_nodes if node.start <= secs < node.end]
        if not query:
            on_note_index = find_index(self.note_list, lambda note: note.start <= secs < note.end)
            if on_note_index >= 0:
                return self.note_list[on_note_index].key * 100
            return (
                min(
                    self.note_list,
                    key=lambda note: min(abs(note.start - secs), abs(secs - note.end)),
                ).key
                * 100
            )
        elif len(query) == 1:
            return query[0].value_at_secs(secs)
        elif len(query) <= 3:
            first = query[0]
            last = query[-1]
            width = first.end - last.start
            bottom = first.value_at_secs(last.start)
            top = last.value_at_secs(first.end)
            diff1 = first.slope_at_secs(last.start)
            diff2 = last.slope_at_secs(first.end)
            return self.cubic_bezier(width, bottom, top, diff1, diff2)(secs - last.start)
        else:
            msg = "More than three sigmoid nodes overlapped"
            raise ParamsError(msg)

    @staticmethod
    def cubic_bezier(
        width: float, bottom: float, top: float, diff1: float, diff2: float
    ) -> Callable[[float], float]:
        a = (2 * (bottom - top) + (diff1 + diff2) * width) / width**3
        b = (3 * (top - bottom) - 2 * diff1 * width - diff2 * width) / width**2
        c = diff1
        d = bottom
        return lambda x: a * x**3 + b * x**2 + c * x + d


@dataclasses.dataclass
class GaussianNode:
    start: float = dataclasses.field(init=False)
    end: float = dataclasses.field(init=False)
    origin: float = dataclasses.field(init=False)
    gaussian_l: Callable[[float], float] = dataclasses.field(init=False)
    gaussian_r: Callable[[float], float] = dataclasses.field(init=False)
    is_end_point: dataclasses.InitVar[bool]
    _origin: dataclasses.InitVar[float]
    _depth: dataclasses.InitVar[float]
    _width: dataclasses.InitVar[float]
    _length_l: dataclasses.InitVar[float]
    _length_r: dataclasses.InitVar[float]
    ratio_miu: float = 0.447684
    ratio_sigma: float = 0.415
    expand: float = 2.5

    def __post_init__(
        self,
        is_end_point: bool,
        _origin: float,
        _depth: float,
        _width: float,
        _length_l: float,
        _length_r: float,
    ) -> None:
        self.origin = _origin
        _depth *= 100
        sigma_base = abs(self.ratio_sigma * _width)
        if is_end_point:
            sigma_l = min(sigma_base, _length_l / self.expand)
            self.start = self.origin - sigma_l * self.expand
            self.gaussian_l = lambda x: _depth * math.exp(-(((x - self.origin) / sigma_l) ** 2))
            sigma_r = min(sigma_base, _length_r / self.expand)
            self.end = self.origin + sigma_r * self.expand
            self.gaussian_r = lambda x: _depth * math.exp(-(((x - self.origin) / sigma_r) ** 2))
        else:
            sign = 1 if _depth > 0 else -1 if _depth < 0 else 0
            depth = abs(_depth)
            miu_base = self.ratio_miu * _width
            r2 = sigma_base**2

            length_base_l = self.expand * sigma_base - miu_base
            if _length_l >= length_base_l:
                self.start = self.origin - length_base_l
                self.gaussian_l = (
                    lambda x: sign
                    * depth
                    * math.exp(-(((x - self.origin - miu_base) / sigma_base) ** 2))
                )
            else:
                self.start = self.origin - _length_l
                k_l = _length_l / length_base_l
                miu_l = miu_base * k_l
                sigma_2l = r2 * k_l
                depth_l = depth * math.exp(miu_l**2 / r2 * (k_l - 1))
                self.gaussian_l = (
                    lambda x: sign
                    * depth_l
                    * math.exp(-((x - self.origin - miu_l) ** 2) / sigma_2l)
                )

            length_base_r = self.expand * sigma_base + miu_base
            if _length_r >= length_base_r:
                self.end = self.origin + length_base_r
                self.gaussian_r = (
                    lambda x: sign
                    * depth
                    * math.exp(-(((x - self.origin - miu_base) / sigma_base) ** 2))
                )
            else:
                self.end = self.origin + _length_r
                k_r = _length_r / length_base_r
                miu_r = miu_base * k_r
                sigma_2r = r2 * k_r
                depth_r = depth * math.exp(miu_r**2 / r2 * (k_r - 1))
                self.gaussian_r = (
                    lambda x: sign
                    * depth_r
                    * math.exp(-((x - self.origin - miu_r) ** 2) / sigma_2r)
                )

    def value_at_secs(self, secs: float) -> float:
        return self.gaussian_l(secs) if secs < self.origin else self.gaussian_r(secs)


@dataclasses.dataclass
class GaussianLayerGenerator:
    _note_list: dataclasses.InitVar[list[NoteStruct]]
    gaussian_nodes: list[GaussianNode] = dataclasses.field(default_factory=list)

    def __post_init__(self, _note_list: list[NoteStruct]) -> None:
        if not _note_list:
            return
        current_note = _note_list[0]
        self.gaussian_nodes.append(
            GaussianNode(
                is_end_point=True,
                _origin=current_note.start,
                _depth=-current_note.depth_left,
                _width=current_note.portamento_left,
                _length_l=math.nan,
                _length_r=current_note.end - current_note.start,
            )
        )
        for current_note, next_note in more_itertools.pairwise(_note_list):
            if next_note.start - current_note.end >= MAX_BREAK:
                self.gaussian_nodes.append(
                    GaussianNode(
                        is_end_point=True,
                        _origin=current_note.end,
                        _depth=current_note.depth_right,
                        _width=current_note.portamento_left,
                        _length_l=current_note.end - current_note.start,
                        _length_r=math.nan,
                    )
                )
                self.gaussian_nodes.append(
                    GaussianNode(
                        is_end_point=True,
                        _origin=next_note.start,
                        _depth=-next_note.depth_left,
                        _width=next_note.portamento_left,
                        _length_l=math.nan,
                        _length_r=next_note.end - next_note.start,
                    )
                )
            else:
                middle = (current_note.end + next_note.start) / 2
                origin = clamp(
                    middle + current_note.portamento_offset,
                    current_note.start,
                    current_note.end,
                )

                depth_l = (
                    current_note.depth_right
                    if current_note.key >= next_note.key
                    else -current_note.depth_right
                )
                self.gaussian_nodes.append(
                    GaussianNode(
                        is_end_point=False,
                        _origin=origin,
                        _depth=depth_l,
                        _width=-current_note.portamento_right,
                        _length_l=origin - current_note.start,
                        _length_r=next_note.end - origin,
                    )
                )

                depth_r = (
                    -next_note.depth_left
                    if current_note.key >= next_note.key
                    else next_note.depth_left
                )
                self.gaussian_nodes.append(
                    GaussianNode(
                        is_end_point=False,
                        _origin=origin,
                        _depth=depth_r,
                        _width=next_note.portamento_left,
                        _length_l=origin - current_note.start,
                        _length_r=next_note.end - origin,
                    )
                )
        self.gaussian_nodes.append(
            GaussianNode(
                is_end_point=True,
                _origin=_note_list[-1].end,
                _depth=-_note_list[-1].depth_right,
                _width=_note_list[-1].portamento_left,
                _length_l=_note_list[-1].end - _note_list[-1].start,
                _length_r=math.nan,
            )
        )

    def pitch_diff_at_secs(self, secs: float) -> float:
        return sum(
            node.value_at_secs(secs)
            for node in self.gaussian_nodes
            if node.start <= secs < node.end
        )


@dataclasses.dataclass
class VibratoNode:
    start: float
    end: float
    fade_left: float
    fade_right: float
    amplitude: float
    frequency: float
    phase: float

    def value_at_secs(self, secs: float) -> float:
        if self.end - self.start >= self.fade_left + self.fade_left:
            if secs < self.start + self.fade_left:
                zoom = (secs - self.start) / self.fade_left
            elif secs > self.end - self.fade_right:
                zoom = (self.end - secs) / self.fade_right
            else:
                zoom = 1.0
        else:
            mid = (self.start * self.fade_right + self.end * self.fade_left) / (
                self.fade_left + self.fade_right
            )
            if secs < mid:
                zoom = (secs - self.start) / self.fade_left
            else:
                zoom = (self.end - secs) / self.fade_right
        return (
            self.amplitude
            * zoom
            * 100.0
            * math.sin(math.tau * self.frequency * (secs - self.start) + self.phase)
        )


@dataclasses.dataclass
class VibratoLayerGenerator:
    _note_list: dataclasses.InitVar[list[NoteStruct]]
    vibrato_nodes: list[VibratoNode] = dataclasses.field(default_factory=list)

    def __post_init__(self, _note_list: list[NoteStruct]) -> None:
        for i in range(len(_note_list)):
            start, end = _note_list[i].start, _note_list[i].end
            if end - start <= _note_list[i].vibrato_start:
                continue
            if i < len(_note_list) - 1 and _note_list[i + 1].start - end < MAX_BREAK:
                end += min(
                    _note_list[i + 1].portamento_offset,
                    min(
                        _note_list[i + 1].vibrato_start,
                        _note_list[i + 1].end - _note_list[i + 1].start,
                    ),
                )
            if start >= end:
                continue
            self.vibrato_nodes.append(
                VibratoNode(
                    start=start + _note_list[i].vibrato_start,
                    end=end,
                    fade_left=_note_list[i].vibrato_left,
                    fade_right=_note_list[i].vibrato_right,
                    amplitude=_note_list[i].vibrato_depth / 2,
                    frequency=_note_list[i].vibrato_frequency,
                    phase=_note_list[i].vibrato_phase,
                )
            )

    def pitch_diff_at_secs(self, secs: float) -> float:
        return sum(
            node.value_at_secs(secs) for node in self.vibrato_nodes if node.start <= secs < node.end
        )


@dataclasses.dataclass
class PitchControlLayerGenerator:
    _pitch_controls: dataclasses.InitVar[list[SVPitchControl]]
    interval_dict: PiecewiseIntervalDict = dataclasses.field(default_factory=PiecewiseIntervalDict)

    def __post_init__(self, _pitch_controls: list[SVPitchControl]) -> None:
        for pitch_control in _pitch_controls:
            if isinstance(pitch_control, SVCurvePitchControl) and len(pitch_control.points):
                prev_point = pitch_control.points.root[0]
                base_position = position_to_ticks(pitch_control.pos)
                prev_ticks = position_to_ticks(prev_point.offset) + base_position
                prev_value = (pitch_control.pitch + prev_point.value) * 100
                self.interval_dict[portion.singleton(prev_ticks)] = prev_value
                for point in pitch_control.points.root[1:]:
                    ticks = position_to_ticks(point.offset) + base_position
                    value = (pitch_control.pitch + point.value) * 100
                    self.interval_dict[
                        portion.openclosed(
                            prev_ticks,
                            ticks,
                        )
                    ] = functools.partial(
                        linear_interpolation,  # type: ignore[call-arg]
                        start=(prev_ticks, prev_value),
                        end=(ticks, value),
                    )
                    prev_ticks = ticks
                    prev_value = value

    def value_at_ticks(self, ticks: int) -> float | None:
        return self.interval_dict.get(ticks)



import dataclasses
import datetime

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Project, SingingTrack
from libresvip.utils.text import LATIN_ALPHABET, SYMBOL_PATTERN

from .model import (
    AlbumInfoTag,
    ArtistInfoTag,
    ByInfoTag,
    LrcFile,
    LyricLine,
    OffsetInfoTag,
    TimeTag,
    TitleInfoTag,
)
from .options import OffsetPolicyOption, OutputOptions, SplitOption


@dataclasses.dataclass
class LrcGenerator:
    options: OutputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> LrcFile:
        self.synchronizer = TimeSynchronizer(project.song_tempo_list)
        singing_track = next(
            track for track in project.track_list if isinstance(track, SingingTrack)
        )
        note_list = singing_track.note_list
        buffer = []
        lyric_lines: list[LyricLine] = []
        for i, note in enumerate(note_list):
            buffer.append((note.start_pos, note.lyric))
            commit_flag = False
            condition_symbol = SYMBOL_PATTERN.search(note.lyric) is not None
            condition_gap = (
                i + 1 < len(note_list) and note_list[i + 1].start_pos - note.end_pos >= 60
            )
            if self.options.split_by == SplitOption.SYMBOL:
                commit_flag = condition_symbol
            elif self.options.split_by == SplitOption.GAP:
                commit_flag = condition_gap
            elif self.options.split_by == SplitOption.BOTH:
                commit_flag = condition_symbol or condition_gap
            if i + 1 == len(note_list):
                commit_flag = True
            if commit_flag:
                self.commit_current_lyric_line(lyric_lines, buffer)
                buffer = []
        info_tags = []
        if self.options.title:
            info_tags.append(TitleInfoTag(value=self.options.title))
        if self.options.artist:
            info_tags.append(ArtistInfoTag(value=self.options.artist))
        if self.options.album:
            info_tags.append(AlbumInfoTag(value=self.options.album))
        if self.options.by:
            info_tags.append(ByInfoTag(value=self.options.by))
        if self.options.offset_policy == OffsetPolicyOption.TIMELINE:
            for line in lyric_lines:
                for time_tag in line.time_tags:
                    ori_time = datetime.datetime(
                        year=1970,
                        month=1,
                        day=1,
                        minute=time_tag.minute,
                        second=time_tag.second,
                        microsecond=time_tag.milisecond * 1000,
                        tzinfo=datetime.timezone.utc,
                    )
                    ori_time += datetime.timedelta(microseconds=-self.options.offset)
                    time_tag.minute = ori_time.minute
                    time_tag.second = ori_time.second
                    time_tag.milisecond = ori_time.microsecond // 1000
        elif self.options.offset_policy == OffsetPolicyOption.META:
            info_tags.append(OffsetInfoTag(value=str(self.options.offset)))
        return LrcFile(
            lyric_lines=lyric_lines,
            info_tags=info_tags,
        )

    def lyric_included(self, lyric: str) -> bool:
        if self.options.ignore_slur_notes:
            return lyric != "-"
        else:
            return True

    def commit_current_lyric_line(
        self, lyric_lines: list[LyricLine], buffer: list[tuple[int, str]]
    ) -> None:
        start_time = self.get_time_from_ticks(buffer[0][0])
        lyrics = ""
        for _, lyric in buffer:
            if self.lyric_included(lyric):
                lyrics += SYMBOL_PATTERN.sub("", lyric) + (
                    " " if LATIN_ALPHABET.search(lyric) is not None else ""
                )
        lyric_lines.append(
            LyricLine(
                time_tags=[
                    TimeTag(
                        minute=start_time.minute,
                        second=start_time.second,
                        milisecond=start_time.microsecond // 1000,
                    )
                ],
                lyric=lyrics + "\n",
            )
        )

    def get_time_from_ticks(self, ticks: int) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            self.synchronizer.get_actual_secs_from_ticks(ticks),
            tz=datetime.timezone.utc,
        )



import io
import pathlib

from libresvip.core.compat import ZipFile, json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis

from .model import VOXFactoryProject
from .options import InputOptions, OutputOptions
from .vox_factory_generator import VOXFactoryGenerator
from .vox_factory_parser import VOXFactoryParser


class VOXFactoryConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        with ZipFile(io.BytesIO(path.read_bytes()), "r") as archive_file:
            proj = VOXFactoryProject.model_validate_json(
                archive_file.read("project.json"),
                context={
                    "extract_audio": options.extract_audio,
                    "path": path,
                    "archive_file": archive_file,
                },
            )
            return VOXFactoryParser(options, path).parse_project(proj)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        if len(project.song_tempo_list) != 1:
            project = reset_time_axis(project, options.tempo)
        buffer = io.BytesIO()
        generator = VOXFactoryGenerator(options)
        vox_factory_project = generator.generate_project(project)
        with ZipFile(buffer, "w") as archive_file:
            archive_file.writestr(
                "project.json",
                json.dumps(
                    vox_factory_project.model_dump(mode="json", by_alias=True),
                    ensure_ascii=False,
                ),
            )
            archive_file.mkdir("resources")
            for audio_name, audio_path in generator.audio_paths.items():
                archive_file.writestr(f"resources/{audio_name}", audio_path.read_bytes())
        path.write_bytes(buffer.getvalue())



from typing import Annotated

import typer

from libresvip.core.config import save_settings, settings
from libresvip.utils.translation import gettext_lazy as _

app = typer.Typer()


@app.command("list")
def list_configurations() -> None:
    for name, value in settings.model_dump().items():
        typer.echo(f"{name}: {value}")


def conf_key_callback(value: str) -> str:
    if value not in ["language"]:
        raise typer.BadParameter(
            _("Setting {} is not supported in cli mode.").format(value),
        )
    return value


@app.command("set")
def set_configuration(
    name: Annotated[str, typer.Argument(callback=conf_key_callback)],
    value: Annotated[str, typer.Argument()],
) -> None:
    setattr(settings, name, value)
    save_settings()
    typer.echo(_("Set {} to {}").format(name, value))



from libresvip.core.time_interval import RangeInterval
from libresvip.model.base import Note, ParamCurve, Params, SingingTrack
from libresvip.model.point import Point
from libresvip.utils.search import find_index, find_last_index


def param_curve_override_with(
    main_curve: ParamCurve,
    override_curve: ParamCurve,
    start: int,
    end: int,
    termination: int = 0,
) -> None:
    inserted_points: list[Point] = []
    main_left_index = find_last_index(main_curve.points.root, lambda point: point.x <= start)
    main_right_index = find_index(main_curve.points.root, lambda point: point.x > end)
    override_left_index = find_last_index(
        override_curve.points.root, lambda point: point.x <= start
    )
    override_right_index = find_index(override_curve.points.root, lambda point: point.x > end)
    main_cure_points_count = len(main_curve.points)
    main_left_defined = 0 <= main_left_index < main_cure_points_count - 1 and (
        main_curve.points[main_left_index].x != termination
        and main_curve.points[main_left_index + 1].x != termination
    )
    main_right_defined = 0 < main_right_index <= main_cure_points_count - 1 and (
        main_curve.points[main_right_index - 1].x != termination
        and main_curve.points[main_right_index].x != termination
    )
    override_cure_points_count = len(override_curve.points)
    override_left_defined = 0 <= override_left_index < override_cure_points_count - 1 and (
        override_curve.points[override_left_index].x != termination
        and override_curve.points[override_left_index + 1].x != termination
    )
    override_right_defined = 0 < override_right_index <= override_cure_points_count - 1 and (
        override_curve.points[override_right_index - 1].x != termination
        and override_curve.points[override_right_index].x != termination
    )
    if main_left_defined:
        r = (start - main_curve.points[main_left_index].x) / (
            main_curve.points[main_left_index + 1].x - main_curve.points[main_left_index].x
        )
        inserted_points.append(
            Point(
                start,
                round(
                    r
                    * (
                        main_curve.points[main_left_index + 1].y
                        + main_curve.points[main_left_index].y
                    )
                    * (1 - r)
                ),
            )
        )
    if main_left_defined ^ override_left_defined:
        inserted_points.append(Point(start, termination))
    if override_left_defined:
        r = (start - override_curve.points[override_left_index].x) / (
            override_curve.points[override_left_index + 1].x
            - override_curve.points[override_left_index].x
        )
        inserted_points.append(
            Point(
                start,
                round(
                    r
                    * (
                        override_curve.points[override_left_index + 1].y
                        + override_curve.points[override_left_index].y
                    )
                    * (1 - r)
                ),
            )
        )
    inserted_points.extend(
        override_curve.points[i] for i in range(override_left_index + 1, override_right_index)
    )
    if override_right_defined:
        r = (end - override_curve.points[override_right_index - 1].x) / (
            override_curve.points[override_right_index].x
            - override_curve.points[override_right_index - 1].x
        )
        inserted_points.append(
            Point(
                end,
                round(
                    r
                    * (
                        override_curve.points[override_right_index].y
                        + override_curve.points[override_right_index - 1].y
                    )
                    * (1 - r)
                ),
            )
        )
    if main_right_defined ^ override_right_defined:
        inserted_points.append(Point(end, termination))
    if main_right_defined:
        r = (end - main_curve.points[main_right_index - 1].x) / (
            main_curve.points[main_right_index].x - main_curve.points[main_right_index - 1].x
        )
        inserted_points.append(
            Point(
                end,
                round(
                    r
                    * (
                        main_curve.points[main_right_index].y
                        + main_curve.points[main_right_index - 1].y
                    )
                    * (1 - r)
                ),
            )
        )
    main_curve.points.root = [
        point
        for i, point in enumerate(main_curve.points.root)
        if main_left_index < i < main_right_index
    ]
    main_curve.points.root = (
        main_curve.points.root[:main_left_index]
        + inserted_points
        + main_curve.points.root[main_left_index:]
    )


def params_override_with(
    main_params: Params, override_params: Params, start: int, end: int
) -> None:
    param_curve_override_with(main_params.pitch, override_params.pitch, start, end)
    param_curve_override_with(main_params.volume, override_params.volume, start, end)
    param_curve_override_with(main_params.breath, override_params.breath, start, end)
    param_curve_override_with(main_params.gender, override_params.gender, start, end)
    param_curve_override_with(main_params.strength, override_params.strength, start, end)


def track_override_with(
    track: SingingTrack,
    note_list: list[Note],
    params: Params,
    first_bar_tick: int,
) -> None:
    main_note_list = track.note_list
    interval = RangeInterval()
    main_left_index = main_right_index = -1
    for i in range(len(note_list)):
        main_left_index = (
            find_last_index(
                main_note_list,
                lambda note: note.start_pos <= note_list[i].start_pos,
            )
            if main_left_index < len(main_note_list) - 1
            else -1
        )
        main_left_note = main_note_list[main_left_index] if main_left_index >= 0 else None
        start = (
            note_list[i].start_pos - 120
            if main_left_note is None or main_left_note.end_pos < note_list[i].start_pos - 120
            else (main_left_note.end_pos + note_list[i].start_pos) // 2
        )
        while i < len(note_list) - 1 and note_list[i].end_pos >= note_list[i + 1].start_pos:
            i += 1
        main_right_index = (
            find_index(
                main_note_list[main_left_index:],
                lambda note: note.start_pos > note_list[i].start_pos,
            )
            if main_right_index < len(main_note_list) - 1
            else -1
        )
        main_right_note = main_note_list[main_right_index] if main_right_index >= 0 else None
        end = (
            note_list[i].end_pos + 120
            if main_right_note is None or main_right_note.start_pos > note_list[i].end_pos + 120
            else (main_right_note.start_pos + note_list[i].end_pos) // 2
        )
        interval |= RangeInterval([(start, end)])
    track.note_list.extend(note_list)
    track.note_list.sort(key=lambda note: note.start_pos)
    for start, end in (interval >> first_bar_tick).sub_ranges():
        params_override_with(track.edited_params, params, start, end)



import argparse
import asyncio
import dataclasses
import enum
import functools
import io
import math
import pathlib
import platform
import re
import secrets
import shutil
import textwrap
import traceback
import webbrowser
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from importlib.resources import as_file
from operator import not_
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    SupportsFloat,
    TypeVar,
    get_args,
    get_type_hints,
)
from urllib.parse import quote, unquote

import aiofiles
import more_itertools
from nicegui import app, binding, ui
from nicegui.context import context
from nicegui.elements.switch import Switch
from nicegui.events import (
    GenericEventArguments,
    KeyEventArguments,
    UploadEventArguments,
    ValueChangeEventArguments,
)
from nicegui.storage import request_contextvar
from pydantic import RootModel, create_model
from pydantic.config import JsonValue
from pydantic_core import PydanticUndefined
from pydantic_extra_types.color import Color
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response
from typing_extensions import ParamSpec
from upath import UPath

import libresvip
from libresvip.core.compat import ZipFile
from libresvip.core.config import (
    ConversionMode,
    DarkMode,
    Language,
    LibreSvipBaseUISettings,
    LyricsReplacement,
    LyricsReplaceMode,
    ui_settings_ctx,
)
from libresvip.core.constants import app_dir, res_dir
from libresvip.core.warning_types import CatchWarnings
from libresvip.extension.manager import (
    get_translation,
    middleware_manager,
    plugin_manager,
)
from libresvip.model.base import BaseComplexModel, Project
from libresvip.utils.search import find_index
from libresvip.utils.text import shorten_error_message, supported_charset_names, uuid_str
from libresvip.utils.translation import gettext_lazy as _
from libresvip.utils.translation import lazy_translation
from libresvip.web.elements import QFab, QFabAction

if TYPE_CHECKING:
    from nicegui.elements.select import Select

binding.MAX_PROPAGATION_TIME = 0.03


P = ParamSpec("P")
R = TypeVar("R")


class LibreSvipWebUserSettings(LibreSvipBaseUISettings):
    def __post_init__(self) -> None:
        self.lyric_replace_rules.setdefault("default", [])
        detected_language = None
        request = request_contextvar.get()
        if accept_lang := request.headers.get("Accept-Language"):
            first_lang = accept_lang.split(",")[0].partition(";")[0]
            if first_lang.startswith("zh"):
                detected_language = Language.from_locale("zh_CN")
            elif first_lang.startswith("ja"):
                detected_language = Language.from_locale("ja_JP")
            elif first_lang.startswith("de"):
                detected_language = Language.from_locale("de_DE")
        if detected_language is not None:
            self.language = detected_language


def dark_mode2str(mode: DarkMode) -> bool | None:
    if mode == DarkMode.LIGHT:
        return False
    elif mode == DarkMode.DARK:
        return True


def str2dark_mode(value: bool | None) -> DarkMode:
    return {
        True: DarkMode.DARK,
        False: DarkMode.LIGHT,
        None: DarkMode.SYSTEM,
    }.get(value, DarkMode.SYSTEM)


def int_validator(value: SupportsFloat | None) -> bool:
    if isinstance(value, int):
        return True
    elif isinstance(value, str):
        return value.replace("+", "-").removeprefix("-").isdigit()
    elif isinstance(value, float):
        return value.is_integer()
    else:
        return False


def float_validator(value: SupportsFloat | None) -> bool:
    if isinstance(value, SupportsFloat):
        return not math.isnan(float(value))
    else:
        return False


@dataclasses.dataclass
class ConversionTask:
    name: str
    upload_path: pathlib.Path
    output_path: pathlib.Path
    running: bool
    success: bool | None
    error: str | None
    warning: str | None

    def reset(self) -> None:
        self.running = False
        self.success = None
        self.error = None
        self.warning = None
        if self.output_path.exists():
            if self.output_path.is_dir():
                self.output_path.rmdir()
            else:
                self.output_path.unlink()

    def __del__(self) -> None:
        if app.native.main_window is None and self.upload_path.exists():
            self.upload_path.unlink()
        if self.output_path.exists():
            if self.output_path.is_dir():
                self.output_path.rmdir()
            else:
                self.output_path.unlink()


def export_all(request: Request) -> Response:
    if selected_formats := getattr(
        app.state, f"{request.path_params['client_id']}_selected_formats"
    ):
        return selected_formats.export_all(request)
    else:
        return Response(
            "No selected formats",
            status_code=400,
        )


app.add_route("/export/{client_id}/", export_all, methods=["GET"])


def export_one(request: Request) -> Response:
    if selected_formats := getattr(
        app.state, f"{request.path_params['client_id']}_selected_formats"
    ):
        return selected_formats.export_one(request)
    else:
        return Response(
            "No selected formats",
            status_code=400,
        )


app.add_route("/export/{client_id}/{filename}", export_one, methods=["GET"])

plugin_details = {
    identifier: {
        "name": plugin.name,
        "author": plugin.author,
        "website": plugin.website,
        "description": plugin.description,
        "version": str(plugin.version),
        "suffix": f"(*.{plugin.suffix})",
        "file_format": plugin.file_format,
        "icon_base64": plugin.icon_base64,
    }
    for identifier, plugin in plugin_manager.plugin_registry.items()
}


@ui.page("/")
@ui.page("/?lang={lang}")
def page_layout(lang: str | None = None) -> None:
    settings: LibreSvipBaseUISettings

    if app.native.main_window is not None:
        from libresvip.core.config import save_settings, settings

        if lang is not None:
            settings.language = Language.from_locale(lang)

        app.on_shutdown(save_settings)
    else:
        settings = LibreSvipWebUserSettings()
        settings.lyric_replace_rules.setdefault("default", [])
        session_id = None
        if (request := request_contextvar.get()) is not None:
            session_id = request.session["id"]

        for key, value in app.storage.user.items():
            if hasattr(settings, key):
                if key == "dark_mode":
                    value = str2dark_mode(value)
                elif key == "language":
                    value = Language.from_locale(value)
                setattr(settings, key, value)

        def save_settings() -> None:
            if session_id in app.storage._users:
                storage = app.storage._users[session_id]
                default_settings_dict = RootModel[LibreSvipWebUserSettings](settings).model_dump(
                    mode="json"
                )

                for key, default_value in default_settings_dict.items():
                    if key == "dark_mode":
                        default_value = dark_mode2str(default_value)
                    if key not in storage or storage[key] != default_value:
                        storage[key] = default_value

        context.client.on_disconnect(save_settings)

    translation = get_translation()

    def set_context_vars() -> None:
        lazy_translation.set(translation)
        if app.native.main_window is not None:
            ui_settings_ctx.set(settings)

    def context_vars_wrapper(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            set_context_vars()
            return func(*args, **kwargs)

        return wrapper

    set_context_vars()

    @context_vars_wrapper
    def plugin_info(attr_name: str) -> None:
        attr = getattr(selected_formats, attr_name)
        with ui.row().classes("w-full h-full"):
            with ui.element("div").classes("w-100 h-100") as icon:
                icon.props["style"] = (
                    f"""background: url('data:image/png;base64,{plugin_details[attr]["icon_base64"]}'); background-size: contain; border-radius: 50%; width: 100px; height: 100px"""
                )
            ui.separator().props("vertical")
            with ui.column().classes("justify-center flex-grow"):
                ui.label(_(plugin_details[attr]["name"] or "")).classes(
                    "text-h5 w-full font-bold text-center",
                )
                with ui.row().classes("w-full"):
                    with ui.element("q-chip").props("icon=bookmark").tooltip(_("Version")):
                        ui.label(plugin_details[attr]["version"] or "")
                    ui.separator().props("vertical")
                    with (
                        ui.element("q-chip")
                        .props("icon=person")
                        .tooltip(plugin_details[attr]["website"] or ""),
                        ui.row().classes("items-center"),
                    ):
                        ui.label(_("Author") + ": ")
                        ui.link(
                            _(plugin_details[attr]["author"] or ""),
                            plugin_details[attr]["website"] or "",
                            new_tab=True,
                        )
                        ui.icon("open_in_new")
                with ui.element("q-chip").props("icon=outline_insert_drive_file"):
                    ui.label(
                        _(plugin_details[attr]["file_format"] or "")
                        + " "
                        + (plugin_details[attr]["suffix"] or ""),
                    )
        ui.separator()
        with ui.card_section().classes("w-full"):
            ui.label(_("Introduction")).classes("text-subtitle1 font-bold")
            ui.label(_(plugin_details[attr]["description"] or ""))

    input_plugin_info = ui.refreshable(functools.partial(plugin_info, "input_format"))
    output_plugin_info = ui.refreshable(functools.partial(plugin_info, "output_format"))

    def panel_header(attr_name: str, title: str, prefix: str, icon: str) -> None:
        attr = getattr(selected_formats, attr_name)
        with ui.row().classes("w-full items-center"):
            ui.icon(icon).classes("text-lg")
            ui.label(title).classes("text-subtitle1 font-bold")
            ui.label(prefix + _(plugin_details[attr]["file_format"] or "") + "]").classes(
                "flex-grow",
            )

    input_panel_header = ui.refreshable(
        functools.partial(
            panel_header,
            "input_format",
            _("Input Options"),
            _("[Import as "),
            "input",
        ),
    )
    output_panel_header = ui.refreshable(
        functools.partial(
            panel_header,
            "output_format",
            _("Output Options"),
            _("[Export to "),
            "output",
        ),
    )

    @context_vars_wrapper
    def options_form(attr_prefix: str, method: str) -> None:
        attr = getattr(selected_formats, attr_prefix + "_format")
        conversion_plugin = plugin_manager.plugin_registry[attr]
        option_class = None
        if hasattr(conversion_plugin.plugin_object, method) and (
            _option_class := get_type_hints(
                getattr(conversion_plugin.plugin_object, method),
            ).get("options")
        ):
            option_class = _option_class
        if not option_class:
            return
        option_dict = getattr(selected_formats, attr_prefix + "_options")
        option_dict.clear()
        option_dict.update(option_class().model_dump())
        with ui.column().classes("w-full"):
            for i, (option_key, field_info) in enumerate(
                option_class.model_fields.items(),
            ):
                default_value = (
                    None if field_info.default is PydanticUndefined else field_info.default
                )
                with ui.row().classes("items-center w-full") as row:
                    if i:
                        row.style("""
                            background-image: linear-gradient(to right, #ccc 0%, #ccc 50%, transparent 50%);
                            background-size: 8px 1px;
                            background-repeat: repeat-x
                        """)
                    if issubclass(field_info.annotation, bool):
                        ui.switch(
                            _(field_info.title),
                            value=default_value,
                        ).bind_value(
                            option_dict,
                            option_key,
                        ).classes("flex-grow")
                    elif issubclass(field_info.annotation, enum.Enum):
                        annotations = get_type_hints(
                            field_info.annotation,
                            include_extras=True,
                        )
                        choices = {}
                        for enum_item in field_info.annotation:
                            if enum_item.name in annotations:
                                annotated_args = list(
                                    get_args(annotations[enum_item.name]),
                                )
                                if len(annotated_args) >= 2:
                                    enum_field = annotated_args[1]
                                else:
                                    continue
                                choices[enum_item] = _(enum_field.title)
                        ui.select(
                            choices,
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif option_key in ["encoding", "lyric_encoding"]:
                        choices = {charset: charset for charset in supported_charset_names()}
                        ui.select(
                            choices,
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, Color):
                        ui.color_input(
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, str | BaseComplexModel):
                        if issubclass(field_info.annotation, BaseComplexModel):
                            default_value = field_info.annotation.default_repr()
                            option_dict[option_key] = default_value
                        ui.input(
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, int | float):
                        with (
                            ui.number(
                                label=_(field_info.title),
                                value=default_value,
                            )
                            .bind_value(option_dict, option_key)
                            .classes(
                                "flex-grow",
                            ) as num_input
                        ):
                            if issubclass(field_info.annotation, int):
                                num_input.validation = {
                                    _("Invalid integer"): int_validator,
                                }
                            else:
                                num_input.validation = {
                                    _("Invalid float"): float_validator,
                                }
                    else:
                        continue
                    if field_info.description:
                        ui.icon("help_outline").classes("text-3xl").style(
                            "cursor: help",
                        ).tooltip(_(field_info.description))

    input_options = ui.refreshable(functools.partial(options_form, "input", "load"))

    output_options = ui.refreshable(functools.partial(options_form, "output", "dump"))

    @context_vars_wrapper
    def middleware_options_form(attr: str, toggler: Switch) -> None:
        conversion_plugin = middleware_manager.plugin_registry[attr]
        field_types = {}
        option_class = None
        if (
            hasattr(conversion_plugin.plugin_object, "process")
            and (
                option_class := get_type_hints(
                    getattr(conversion_plugin.plugin_object, "process"),
                ).get("options")
            )
            and hasattr(option_class, "model_fields")
        ):
            for option_key, field_info in option_class.model_fields.items():
                if issubclass(
                    field_info.annotation,
                    str | Color | enum.Enum | BaseComplexModel,
                ):
                    field_types[option_key] = str
                else:
                    field_types[option_key] = field_info.annotation
        if not option_class or not field_types:
            return
        option_dict = getattr(selected_formats.middleware_options, attr)
        with ui.column().classes("w-full").bind_visibility_from(toggler, "value"):
            for i, (option_key, field_info) in enumerate(
                option_class.model_fields.items(),
            ):
                default_value = (
                    None if field_info.default is PydanticUndefined else field_info.default
                )
                with ui.row().classes("items-center w-full") as row:
                    if i:
                        row.style("""
                            background-image: linear-gradient(to right, #ccc 0%, #ccc 50%, transparent 50%);
                            background-size: 8px 1px;
                            background-repeat: repeat-x
                        """)
                    if issubclass(field_info.annotation, bool):
                        ui.switch(
                            _(field_info.title),
                            value=default_value,
                        ).bind_value(
                            option_dict,
                            option_key,
                        ).classes("flex-grow")
                    elif issubclass(field_info.annotation, enum.Enum):
                        annotations = get_type_hints(
                            field_info.annotation,
                            include_extras=True,
                        )
                        choices = {}
                        for enum_item in field_info.annotation:
                            if enum_item.name in annotations:
                                annotated_args = list(
                                    get_args(annotations[enum_item.name]),
                                )
                                if len(annotated_args) >= 2:
                                    enum_field = annotated_args[1]
                                else:
                                    continue
                                choices[enum_item] = _(enum_field.title)
                        ui.select(
                            choices,
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif option_key == "lyric_replacement_preset_name":
                        choices = {preset: preset for preset in settings.lyric_replace_rules}
                        ui.select(
                            choices,
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, Color):
                        ui.color_input(
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, str | BaseComplexModel):
                        if issubclass(field_info.annotation, BaseComplexModel):
                            default_value = field_info.annotation.default_repr()
                            option_dict[option_key] = default_value
                        ui.input(
                            label=_(field_info.title),
                            value=default_value,
                        ).bind_value(option_dict, option_key).classes("flex-grow")
                    elif issubclass(field_info.annotation, int | float):
                        with (
                            ui.number(
                                label=_(field_info.title),
                                value=default_value,
                            )
                            .bind_value(option_dict, option_key)
                            .classes(
                                "flex-grow",
                            ) as num_input
                        ):
                            if issubclass(field_info.annotation, int):
                                num_input.validation = {
                                    _("Invalid integer"): int_validator,
                                }
                            else:
                                num_input.validation = {
                                    _("Invalid float"): float_validator,
                                }
                    else:
                        continue
                    if field_info.description:
                        ui.icon("help_outline").classes("text-3xl").style(
                            "cursor: help",
                        ).tooltip(_(field_info.description))

    select_input: Select
    select_output: Select

    def _input_format_factory() -> str:
        return (
            settings.last_input_format
            if settings.last_input_format is not None
            else next(
                iter(plugin_manager.plugin_registry),
                "",
            )
        )

    def _output_format_factory() -> str:
        return (
            settings.last_output_format
            if settings.last_output_format is not None
            else next(
                iter(plugin_manager.plugin_registry),
                "",
            )
        )

    @dataclasses.dataclass
    class SelectedFormats:
        _input_format: str = dataclasses.field(default_factory=_input_format_factory)
        _output_format: str = dataclasses.field(default_factory=_output_format_factory)
        _conversion_mode: ConversionMode = dataclasses.field(default=ConversionMode.DIRECT)
        current_preset: str = dataclasses.field(default="default")
        input_options: dict[str, JsonValue] = dataclasses.field(default_factory=dict)
        output_options: dict[str, JsonValue] = dataclasses.field(default_factory=dict)
        files_to_convert: dict[str, ConversionTask] = dataclasses.field(
            default_factory=dict,
        )

        def __post_init__(self) -> None:
            self.middleware_enabled_states = create_model(
                "middleware_enabled_states",
                **dict.fromkeys(middleware_manager.plugin_registry, (bool, False)),
            )()
            self.middleware_options = create_model(
                "middleware_options",
                **{
                    abbr: (options, options())
                    for abbr, middleware in middleware_manager.plugin_registry.items()
                    if (
                        middleware.plugin_object is not None
                        and hasattr(middleware.plugin_object, "process")
                        and (
                            options := get_type_hints(middleware.plugin_object.process).get(
                                "options",
                            )
                        )
                    )
                },
            )()

        @functools.cached_property
        def temp_path(self) -> UPath:
            user_temp_path = UPath("memory:/") / f"{context.client.id}"
            if not user_temp_path.exists():
                user_temp_path.mkdir(exist_ok=True)
            return user_temp_path

        def reset(self) -> None:
            self.files_to_convert.clear()
            self.tasks_container.refresh()

        def filter_input_ext(self) -> None:
            self.files_to_convert = {
                name: info
                for name, info in self.files_to_convert.items()
                if info.upload_path.suffix == f".{self.input_format}"
            }
            self.tasks_container.refresh()

        @ui.refreshable
        @context_vars_wrapper
        def tasks_container(self) -> None:
            with ui.scroll_area().classes("w-full"):
                for i, info in enumerate(self.files_to_convert.values()):
                    with ui.row().classes("w-full items-center"):

                        def remove_row() -> None:
                            del self.files_to_convert[info.name]
                            self.tasks_container.refresh()

                        ui.label(info.name).classes("flex-grow")
                        if self._conversion_mode == ConversionMode.MERGE and i:
                            ui.icon("merge", size="lg")
                        ui.spinner().props("size=lg").bind_visibility_from(
                            info,
                            "running",
                        )
                        ui.icon("check", size="lg").classes(
                            "text-green-500",
                        ).bind_visibility_from(info, "success")
                        with (
                            ui.dialog() as error_dialog,
                            ui.element(
                                "q-banner",
                            ).classes("bg-red-500 w-auto") as error_banner,
                        ):
                            with (
                                ui.scroll_area()
                                .classes(
                                    remove="nicegui-scroll-area",
                                )
                                .style("width: 500px; height: 16rem;")
                            ):
                                ui.label().classes("text-lg").style(
                                    "word-break: break-all; white-space: pre-wrap",
                                ).bind_text_from(
                                    info,
                                    "error",
                                    backward=shorten_error_message,
                                )
                            with error_banner.add_slot("action"):
                                ui.button(
                                    _("Copy to clipboard"),
                                    on_click=lambda: ui.clipboard.write(info.error or "")
                                    and ui.notify(_("Copied"), type="info"),
                                )
                                ui.button(_("Close"), on_click=error_dialog.close)
                        ui.button(
                            icon="error",
                            color="red",
                            on_click=error_dialog.open,
                        ).props("round").bind_visibility_from(info, "error")
                        with (
                            ui.dialog() as warn_dialog,
                            ui.element("q-banner").classes(
                                "bg-yellow-500 w-auto",
                            ) as warn_banner,
                        ):
                            with (
                                ui.scroll_area()
                                .classes(
                                    remove="nicegui-scroll-area",
                                )
                                .style("width: 500px; height: 16rem;")
                            ):
                                ui.label().classes("text-lg").style(
                                    "word-break: break-all; white-space: pre-wrap",
                                ).bind_text_from(info, "warning", backward=str)
                            with warn_banner.add_slot("action"):
                                ui.button(
                                    _("Copy to clipboard"),
                                    on_click=lambda: ui.clipboard.write(info.warning or "")
                                    and ui.notify(_("Copied"), type="info"),
                                )
                                ui.button(_("Close"), on_click=warn_dialog.close)
                        ui.button(
                            icon="warning",
                            color="yellow",
                            on_click=warn_dialog.open,
                        ).props("round").bind_visibility_from(info, "warning")
                        ui.button(
                            icon="download",
                            on_click=functools.partial(self.save_file, info.name),
                        ).props("round").bind_visibility_from(info, "success")
                        ui.button(icon="close", on_click=remove_row).props(
                            "round",
                        ).tooltip(_("Remove"))

        @property
        def input_format(self) -> str:
            return self._input_format

        @input_format.setter
        def input_format(self, value: str) -> None:
            if value != self._input_format:
                self._input_format = value
                if settings.reset_tasks_on_input_change:
                    self.files_to_convert.clear()
                settings.last_input_format = value
                input_plugin_info.refresh()
                input_panel_header.refresh()
                input_options.refresh()

        @property
        def output_format(self) -> str:
            return self._output_format

        @output_format.setter
        def output_format(self, value: str) -> None:
            if value != self._output_format:
                self._output_format = value
                settings.last_output_format = value
                for task in self.files_to_convert.values():
                    task.reset()
                output_plugin_info.refresh()
                output_panel_header.refresh()
                output_options.refresh()

        @context_vars_wrapper
        async def _add_task(
            self,
            name: str,
            content: BinaryIO | pathlib.Path,
        ) -> None:
            if settings.auto_detect_input_format:
                cur_suffix = name.rpartition(".")[-1].lower()
                if cur_suffix in plugin_manager.plugin_registry and cur_suffix != self.input_format:
                    self.input_format = cur_suffix
            if isinstance(content, pathlib.Path):
                upload_path = content
            else:
                upload_path = self.temp_path / name
                content.seek(0)
                upload_path.write_bytes(content.read())
            output_path = self.temp_path / uuid_str()
            conversion_task = ConversionTask(
                name=name,
                upload_path=upload_path,
                output_path=output_path,
                running=False,
                success=None,
                error=None,
                warning=None,
            )
            self.files_to_convert[name] = conversion_task
            self.tasks_container.refresh()

        async def add_task(self, args: UploadEventArguments) -> None:
            await self._add_task(args.name, args.content)

        @property
        def conversion_mode(self) -> str:
            return self._conversion_mode.value

        @conversion_mode.setter
        def conversion_mode(self, value: str) -> None:
            if value != self._conversion_mode.value:
                self._conversion_mode = ConversionMode(value)
                for task in self.files_to_convert.values():
                    task.reset()
                self.tasks_container.refresh()

        @property
        def task_count(self) -> int:
            return len(self.files_to_convert)

        @context_vars_wrapper
        def convert_one(self, task: ConversionTask, *sub_tasks: list[ConversionTask]) -> None:
            task.reset()
            task.running = True
            try:
                with CatchWarnings() as w:
                    input_plugin = plugin_manager.plugin_registry[self.input_format]
                    output_plugin = plugin_manager.plugin_registry[self.output_format]
                    if (
                        input_plugin.plugin_object is None
                        or (
                            input_option_class := get_type_hints(
                                input_plugin.plugin_object.load
                            ).get(
                                "options",
                            )
                        )
                        is None
                        or output_plugin.plugin_object is None
                        or (
                            output_option_class := get_type_hints(
                                output_plugin.plugin_object.dump,
                            ).get("options")
                        )
                        is None
                    ):
                        task.success = False
                    else:
                        input_option = input_option_class(**self.input_options)
                        if self._conversion_mode == ConversionMode.MERGE:
                            child_projects = [
                                input_plugin.plugin_object.load(
                                    sub_task.upload_path,
                                    input_option,
                                )
                                for sub_task in more_itertools.value_chain(task, sub_tasks)
                            ]
                            project = Project.merge_projects(child_projects)
                        else:
                            project = input_plugin.plugin_object.load(
                                task.upload_path,
                                input_option,
                            )
                        if self._conversion_mode != ConversionMode.SPLIT:
                            task.output_path = task.output_path.with_suffix(
                                f".{self.output_format}",
                            )
                        for (
                            middleware_abbr,
                            enabled,
                        ) in self.middleware_enabled_states.model_dump().items():
                            if enabled:
                                middleware = middleware_manager.plugin_registry[middleware_abbr]
                                if middleware.plugin_object is not None and hasattr(
                                    middleware.plugin_object, "process"
                                ):
                                    middleware_option = getattr(
                                        self.middleware_options,
                                        middleware_abbr,
                                    )
                                    project = middleware.plugin_object.process(
                                        project,
                                        middleware_option.model_validate(
                                            middleware_option,
                                            from_attributes=True,
                                        ),
                                    )
                        output_option = output_option_class(**self.output_options)
                        if self._conversion_mode == ConversionMode.SPLIT:
                            task.output_path.mkdir(parents=True, exist_ok=True)
                            for i, child_project in enumerate(
                                project.split_tracks(settings.max_track_count)
                            ):
                                output_plugin.plugin_object.dump(
                                    task.output_path
                                    / f"{task.upload_path.stem}_{i + 1:0=2d}.{self.output_format}",
                                    child_project,
                                    output_option,
                                )
                        else:
                            output_plugin.plugin_object.dump(
                                task.output_path,
                                project,
                                output_option,
                            )
                        task.success = True
                if w.output:
                    task.warning = w.output
            except Exception:
                task.success = False
                task.error = traceback.format_exc()
            task.running = False

        @context_vars_wrapper
        async def batch_convert(self) -> None:
            loop = asyncio.get_running_loop()
            running_tasks = []
            with ThreadPoolExecutor(
                max_workers=min(len(self.files_to_convert), 4),
            ) as executor:
                futures: list[asyncio.Future[None]]
                if self._conversion_mode == ConversionMode.MERGE:
                    (task,), other_tasks = more_itertools.spy(self.files_to_convert.values())
                    futures = [
                        loop.run_in_executor(
                            executor,
                            self.convert_one,
                            task,
                            *other_tasks,
                        )
                    ]
                    running_tasks.append(task)
                else:
                    running_tasks = list(self.files_to_convert.values())
                    futures = [
                        loop.run_in_executor(
                            executor,
                            self.convert_one,
                            task,
                        )
                        for task in running_tasks
                    ]
                for future in asyncio.as_completed(futures):
                    await future
            if any(not task.success for task in running_tasks):
                ui.notification(_("Conversion Failed"), type="negative")
            else:
                ui.notification(_("Conversion Successful"), type="positive")

        def export_all(self, request: Request) -> Response:
            if result := self._export_all():
                return Response(*result)
            raise HTTPException(400, "No files to export")

        def export_one(self, request: Request) -> Response:
            if result := self._export_one(request.path_params["filename"]):
                return Response(*result)
            raise HTTPException(404, "File not found")

        def _export_one(self, filename: str) -> tuple[bytes, int, dict[str, str], str] | None:
            if not (task := self.files_to_convert.get(filename)) or not task.success:
                return None
            if self._conversion_mode == ConversionMode.SPLIT:
                buffer = io.BytesIO()
                with ZipFile(buffer, "w") as zip_file:
                    for child_file in task.output_path.iterdir():
                        if not child_file.is_file():
                            continue
                        zip_file.writestr(
                            child_file.name,
                            child_file.read_bytes(),
                        )
                content = buffer.getvalue()
                filename_header = quote(task.upload_path.with_suffix(".zip").name)
            else:
                content = task.output_path.read_bytes()
                filename_header = quote(task.upload_path.with_suffix(task.output_path.suffix).name)
            return (
                content,
                200,
                {
                    "Content-Disposition": f"attachment; filename={filename_header}",
                },
                "application/octet-stream",
            )

        def _export_all(
            self,
        ) -> tuple[bytes, int, dict[str, str], str] | None:
            if len(self.files_to_convert) == 0:
                return None
            elif len(self.files_to_convert) == 1:
                filename = next(iter(self.files_to_convert))
                return self._export_one(filename)
            buffer = io.BytesIO()
            with ZipFile(buffer, "w") as zip_file:
                for task in self.files_to_convert.values():
                    if task.success:
                        if task.output_path.is_dir():
                            for i, child_file in enumerate(task.output_path.iterdir()):
                                if not child_file.is_file():
                                    continue
                                zip_file.writestr(
                                    child_file.name,
                                    child_file.read_bytes(),
                                )
                        else:
                            zip_file.writestr(
                                task.upload_path.with_suffix(task.output_path.suffix).name,
                                task.output_path.read_bytes(),
                            )
            return (
                buffer.getvalue(),
                200,
                {"Content-Disposition": "attachment; filename=export.zip"},
                "application/zip",
            )

        @context_vars_wrapper
        async def add_upload(self) -> None:
            nonlocal select_input
            if app.native.main_window is not None and hasattr(
                app.native.main_window, "create_file_dialog"
            ):
                file_paths = await app.native.main_window.create_file_dialog(
                    allow_multiple=True,
                    file_types=[
                        " " + select_input.options[select_input.value].rpartition(" ")[-1],
                        _("All files (*.*)"),
                    ],
                )
                if file_paths is None:  # Canceled
                    return
                for file_path in file_paths:
                    path = pathlib.Path(file_path)
                    await self._add_task(path.name, path)
            else:
                ui.run_javascript("add_upload()")

        @context_vars_wrapper
        async def save_file(self, file_name: str = "") -> None:
            nonlocal select_output
            if app.native.main_window is not None and hasattr(
                app.native.main_window, "create_file_dialog"
            ):
                import webview

                result = None
                result = self._export_one(file_name) if file_name else self._export_all()
                if result is None:
                    ui.notify(_("Save failed!"), type="negative")
                    return

                save_filename = unquote(
                    result[2]["Content-Disposition"].removeprefix("attachment; filename=")
                )
                save_path = await app.native.main_window.create_file_dialog(
                    webview.SAVE_DIALOG,
                    save_filename=save_filename,
                    file_types=(
                        _("Compressed Archive (*.zip)")
                        if save_filename.endswith(".zip")
                        else " " + select_output.options[select_output.value].rpartition(" ")[-1],
                        _("All files (*.*)"),
                    ),
                )
                if save_path is None:  # Canceled
                    return
                elif not isinstance(save_path, str):  # list[str]
                    save_path = save_path[0]
                async with aiofiles.open(save_path, "wb") as content:
                    await content.write(result[0])
                ui.notify(_("Saved"), type="positive")
            else:
                ui.download(f"/export/{context.client.id}/{file_name}")

    dark_toggler = ui.dark_mode().bind_value(
        settings, "dark_mode", forward=str2dark_mode, backward=dark_mode2str
    )
    dark_toggler.on_value_change(lambda: ui.navigate.to("/"))
    selected_formats = SelectedFormats()
    if app.native.main_window is None:
        setattr(
            app.state,
            f"{context.client.id}_selected_formats",
            selected_formats,
        )

        def recycle_state() -> None:
            delattr(app.state, f"{context.client.id}_selected_formats")

        context.client.on_disconnect(recycle_state)
    ui.add_head_html(
        textwrap.dedent(
            """
        <style>
            .q-icon {
                justify-content: flex-end;
            }
        </style>
        """
        ).strip()
    )  # fix icon position

    def swap_values() -> None:
        selected_formats.input_format, selected_formats.output_format = (
            selected_formats.output_format,
            selected_formats.input_format,
        )

    ui.colors(
        primary="#3F51B5",
        secondary="#5C6BC0",
        accent="#8A72AC",
        dark="#212121",
        positive="#28A745",
        negative="#D32F2F",
        info="#536DFE",
        warning="#FFB74D",
    )
    with (
        ui.header(elevated=True)
        .style("background-color: primary")
        .classes(
            "items-center pywebview-drag-region",
        )
    ):
        with ui.row().classes("w-full"):
            with ui.dialog() as about_dialog, ui.card():
                ui.label(_("About")).classes("text-lg")
                with ui.column().classes("text-center w-full"):
                    ui.label(_("LibreSVIP")).classes("text-h4 font-bold w-full")
                    ui.label(_("Version: ") + libresvip.__version__).classes(
                        "text-md w-full",
                    )
                    ui.label(_("Author: SoulMelody")).classes("text-md w-full")
                    with ui.row().classes("w-full justify-center"):
                        with ui.element("q-chip").props("icon=live_tv"):
                            ui.link(
                                _("Author's Profile"),
                                "https://space.bilibili.com/175862486",
                                new_tab=True,
                            )
                        with ui.element("q-chip").props("icon=logo_dev"):
                            ui.link(
                                _("Repo URL"),
                                "https://github.com/SoulMelody/LibreSVIP",
                                new_tab=True,
                            )
                    ui.label(
                        _(
                            "LibreSVIP is an open-sourced, liberal and extensionable framework that can convert your singing synthesis projects between different file formats.",
                        ),
                    ).classes("text-md w-full")
                    ui.label(
                        _(
                            "All people should have the right and freedom to choose. That's why we're committed to giving you a second chance to keep your creations free from the constraints of platforms and coterie.",
                        ),
                    ).classes("text-md w-full")
                with ui.card_actions().props("align=right").classes("w-full"):
                    ui.button(_("Close"), on_click=about_dialog.close)

            def toggle_menu() -> None:
                submenu.classes(toggle="hidden")
                toggle_menu_btn.props(
                    "icon=arrow_drop_down" if "hidden" in submenu.classes else "icon=arrow_drop_up"
                )

            toggle_menu_btn = (
                ui.button(icon="arrow_drop_down", on_click=toggle_menu)
                .classes("aspect-square")
                .tooltip(_("Toggle menu"))
            )
            with (
                ui.dialog() as settings_dialog,
                ui.card().classes("min-w-[750px]"),
            ):
                with ui.splitter(value=25, limits=(25, 30)).classes("w-full") as settings_splitter:
                    with (
                        settings_splitter.before,
                        ui.tabs().props("vertical") as settings_nav,
                    ):
                        conversion_settings_tab = ui.tab(
                            _("Conversion Settings"),
                            icon="settings_applications",
                        )
                        lyric_replace_rules_tab = ui.tab(
                            _("Lyric Replace Rules"), icon="text_rotation_none"
                        )
                        language_tab = ui.tab(_("Switch Language"), icon="language")
                    with (
                        settings_splitter.after,
                        ui.tab_panels(settings_nav, value=conversion_settings_tab),
                    ):
                        with (
                            ui.tab_panel(conversion_settings_tab),
                            ui.column(),
                        ):
                            ui.switch(_("Auto detect import format")).bind_value(
                                settings,
                                "auto_detect_input_format",
                            )
                            ui.switch(
                                _("Reset list when import format changed"),
                            ).classes("col-span-5").bind_value(
                                settings,
                                "reset_tasks_on_input_change",
                            )
                        with ui.tab_panel(lyric_replace_rules_tab):
                            columns = [
                                {
                                    "name": "mode",
                                    "label": _("Mode"),
                                    "field": "mode",
                                },
                                {
                                    "name": "pattern_prefix",
                                    "label": _("Prefix"),
                                    "field": "pattern_prefix",
                                },
                                {
                                    "name": "pattern_main",
                                    "label": _("Pattern"),
                                    "field": "pattern_main",
                                },
                                {
                                    "name": "pattern_suffix",
                                    "label": _("Suffix"),
                                    "field": "pattern_suffix",
                                },
                                {
                                    "name": "replacement",
                                    "label": _("Replacement"),
                                    "field": "replacement",
                                },
                                {
                                    "name": "flags",
                                    "label": _("Flags"),
                                    "field": "flags",
                                },
                                {
                                    "name": "actions",
                                    "label": _("Actions"),
                                    "field": "actions",
                                },
                            ]

                            def refresh_rules(
                                preset: str,
                            ) -> list[dict[str, JsonValue]]:
                                return [
                                    {
                                        "id": i + 1,
                                        "mode": rule.mode.value,
                                        "pattern_prefix": rule.pattern_prefix,
                                        "pattern_main": rule.pattern_main,
                                        "pattern_suffix": rule.pattern_suffix,
                                        "replacement": rule.replacement,
                                        "flags": rule.flags.value,
                                    }
                                    for i, rule in enumerate(settings.lyric_replace_rules[preset])
                                ]

                            rows = refresh_rules(selected_formats.current_preset)
                            replace_mode_options = [
                                {"label": _("Full match"), "value": "full"},
                                {
                                    "label": _("Alphabetic"),
                                    "value": "alphabetic",
                                },
                                {
                                    "label": _("Non-alphabetic"),
                                    "value": "non_alphabetic",
                                },
                                {"label": _("Regex"), "value": "regex"},
                            ]
                            re_flags_options = [
                                {
                                    "label": _("Ignore case"),
                                    "value": re.IGNORECASE.value,
                                },
                                {
                                    "label": _("Case sensitive"),
                                    "value": re.UNICODE.value,
                                },
                            ]
                            table = ui.table(columns=columns, rows=rows, pagination=5).classes(
                                "w-full"
                            )
                            table.add_slot(
                                "body-cell-mode",
                                r'''
                                <q-td key="mode" :props="props">
                                    <q-select
                                        v-model="props.row.mode"
                                        readonly
                                        map-options
                                        :options="'''
                                + str(replace_mode_options)
                                + r""""
                                    />
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-pattern_prefix",
                                r"""
                                <q-td key="pattern_prefix" :props="props">
                                    {{ props.row.pattern_prefix }}
                                    <q-popup-edit v-model="props.row.pattern_prefix" v-if="props.row.mode === 'regex'" v-slot="scope">
                                        <q-input
                                            v-model="scope.value" dense autofocus counter
                                            @update:model-value="() => {props.row.pattern_prefix = scope.value; $parent.$emit('modify_field', props.row.id, 'pattern_prefix', scope.value)}"
                                        />
                                    </q-popup-edit>
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-pattern_main",
                                r"""
                                <q-td key="pattern_main" :props="props">
                                    {{ props.row.pattern_main }}
                                    <q-popup-edit v-model="props.row.pattern_main" v-slot="scope">
                                        <q-input
                                            v-model="scope.value" dense autofocus counter
                                            @update:model-value="() => {props.row.pattern_main = scope.value; $parent.$emit('modify_field', props.row.id, 'pattern_main', scope.value)}"
                                        />
                                    </q-popup-edit>
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-pattern_suffix",
                                r"""
                                <q-td key="pattern_suffix" :props="props">
                                    {{ props.row.pattern_suffix }}
                                    <q-popup-edit v-model="props.row.pattern_suffix" v-if="props.row.mode === 'regex'" v-slot="scope">
                                        <q-input
                                            v-model="scope.value" dense autofocus counter
                                            @update:model-value="() => {props.row.pattern_suffix = scope.value; $parent.$emit('modify_field', props.row.id, 'pattern_suffix', scope.value)}"
                                        />
                                    </q-popup-edit>
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-replacement",
                                r"""
                                <q-td key="replacement" :props="props">
                                    {{ props.row.replacement }}
                                    <q-popup-edit v-model="props.row.replacement" v-slot="scope">
                                        <q-input
                                            v-model="scope.value" dense autofocus counter
                                            @update:model-value="() => {props.row.replacement = scope.value; $parent.$emit('modify_field', props.row.id, 'replacement', scope.value)}"
                                        />
                                    </q-popup-edit>
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-flags",
                                r'''
                                <q-td key="flags" :props="props">
                                    <q-select
                                        v-model="props.row.flags"
                                        map-options
                                        :options="'''
                                + str(re_flags_options)
                                + r""""
                                        @update:model-value="() => $parent.$emit('modify_field', props.row.id, 'flags', props.row.flags.value)"
                                    />
                                </q-td>
                            """,
                            )
                            table.add_slot(
                                "body-cell-actions",
                                r"""
                                <q-td key="actions" :props="props">
                                    <span>
                                        <q-btn size="sm" color="accent" round dense
                                            @click="() => $parent.$emit('move_up', props.row)"
                                            icon="arrow_upward" />
                                        <q-btn size="sm" color="accent" round dense
                                            @click="() => $parent.$emit('move_down', props.row)"
                                            icon="arrow_downward" />
                                        <q-btn size="sm" color="accent" round dense
                                            @click="() => $parent.$emit('delete', props.row)"
                                            icon="remove" />
                                    </span>
                                </q-td>
                            """,
                            )

                            def modify_field(
                                event: GenericEventArguments,
                            ) -> None:
                                if (
                                    row_idx := find_index(
                                        rows,
                                        lambda row: row["id"] == event.args["id"],
                                    )
                                ) != -1:
                                    rows[row_idx][event.args[1]] = event.args[2]

                            table.on("modify_field", modify_field)

                            def move_up_rule(
                                event: GenericEventArguments,
                            ) -> None:
                                if (
                                    row_idx := find_index(
                                        rows,
                                        lambda row: row["id"] == event.args["id"],
                                    )
                                ) > 0:
                                    rows[row_idx - 1], rows[row_idx] = (
                                        rows[row_idx],
                                        rows[row_idx - 1],
                                    )
                                    table.update_rows(rows)

                            table.on("move_up", move_up_rule)

                            def move_down_rule(
                                event: GenericEventArguments,
                            ) -> None:
                                if (
                                    -1
                                    < (
                                        row_idx := find_index(
                                            rows,
                                            lambda row: row["id"] == event.args["id"],
                                        )
                                    )
                                    < len(rows) - 1
                                ):
                                    rows[row_idx + 1], rows[row_idx] = (
                                        rows[row_idx],
                                        rows[row_idx + 1],
                                    )
                                    table.update_rows(rows)

                            table.on("move_down", move_down_rule)

                            def delete_rule(
                                event: GenericEventArguments,
                            ) -> None:
                                if (
                                    row_idx := find_index(
                                        rows,
                                        lambda row: row["id"] == event.args["id"],
                                    )
                                ) != -1:
                                    table.remove_row(rows[row_idx])

                            table.on("delete", delete_rule)

                            with table.add_slot("top"):

                                def table_toggle_fullscreen() -> None:
                                    table.toggle_fullscreen()
                                    table_fullscreen_btn.props(
                                        "icon=fullscreen_exit"
                                        if table.is_fullscreen
                                        else "icon=fullscreen"
                                    )

                                table_fullscreen_btn = (
                                    ui.button(
                                        icon="fullscreen",
                                        on_click=table_toggle_fullscreen,
                                    )
                                    .props("round")
                                    .tooltip(_("Toggle fullscreen"))
                                )

                                def refresh_rows(
                                    event: ValueChangeEventArguments,
                                ) -> None:
                                    if event.value is not None:
                                        settings.lyric_replace_rules.setdefault(event.value, [])
                                        table.update_rows(refresh_rules(event.value))

                                preset_select = ui.select(
                                    list(settings.lyric_replace_rules),
                                    label=_("Preset: "),
                                    new_value_mode="add-unique",
                                    on_change=refresh_rows,
                                ).bind_value(selected_formats, "current_preset")

                                def delete_preset() -> None:
                                    if preset_select.value != "default":
                                        settings.lyric_replace_rules.pop(preset_select.value)
                                        preset_select.set_options(
                                            list(settings.lyric_replace_rules),
                                            value="default",
                                        )

                                ui.button(icon="remove", on_click=delete_preset).props(
                                    "round"
                                ).tooltip(_("Remove current preset"))

                                def save_preset() -> None:
                                    settings.lyric_replace_rules[preset_select.value] = [
                                        LyricsReplacement(  # type: ignore[call-arg]
                                            mode=LyricsReplaceMode(row["mode"]),
                                            replacement=row["replacement"],
                                            pattern_main=row["pattern_main"],
                                            pattern_prefix=row["pattern_prefix"],
                                            pattern_suffix=row["pattern_suffix"],
                                            flags=re.UNICODE
                                            if row["flags"] == re.UNICODE.value
                                            else re.IGNORECASE,
                                        )
                                        for row in rows
                                    ]
                                    middleware_options.refresh()

                                ui.button(icon="save", on_click=save_preset).props("round").tooltip(
                                    _("Save current preset")
                                )

                                def add_rule(mode_value: str) -> None:
                                    pattern_prefix = pattern_suffix = ""
                                    if mode_value == LyricsReplaceMode.FULL.value:
                                        pattern_prefix = "^"
                                        pattern_suffix = "$"
                                    elif mode_value == LyricsReplaceMode.ALPHABETIC.value:
                                        pattern_prefix = r"(?<=^|\b)"
                                        pattern_suffix = r"(?=$|\b)"
                                    table.add_row(
                                        {
                                            "id": rows[-1]["id"] + 1 if len(rows) else 1,
                                            "mode": mode_value,
                                            "pattern_main": "",
                                            "pattern_prefix": pattern_prefix,
                                            "pattern_suffix": pattern_suffix,
                                            "replacement": "",
                                            "flags": re.IGNORECASE.value,
                                        }
                                    )

                                with ui.dropdown_button(_("Add new rule"), icon="add").tooltip(
                                    _(
                                        "Alphabetic: Applies to alphabetic characters.\nNon-Alphabetic: For non-alphabetic characters and punctuation marks.\nRegex: for advanced users with knowledge of regular expressions."
                                    )
                                ):
                                    for option in replace_mode_options:
                                        ui.item(
                                            option["label"],
                                            on_click=functools.partial(add_rule, option["value"]),
                                        )
                        with ui.tab_panel(language_tab):

                            def switch_language(
                                event: ValueChangeEventArguments,
                            ) -> None:
                                if event.value == Language.CHINESE:
                                    ui.navigate.to("/?lang=zh_CN")
                                elif event.value == Language.ENGLISH:
                                    ui.navigate.to("/?lang=en_US")
                                elif event.value == Language.GERMAN:
                                    ui.navigate.to("/?lang=de_DE")

                            ui.select(
                                {
                                    Language.CHINESE: "简体中文",
                                    Language.ENGLISH: "English",
                                    Language.JAPANESE: "日本語",
                                    Language.GERMAN: "Deutsch",
                                },
                            ).bind_value(
                                settings,
                                "language",
                            ).on_value_change(switch_language)
                with ui.card_actions().props("align=right").classes("w-full"):
                    ui.button(_("Close"), on_click=settings_dialog.close)
            ui.space()
            with ui.tabs().classes("sm:visible lg:h-0 lg:invisible") as tabs:
                format_select_tab = ui.tab(_("Select File Formats"))
                options_tab = ui.tab(_("Advanced Settings"))
            ui.space()
            if app.native.main_window is not None:
                if platform.system() == "Windows":
                    import ctypes.wintypes

                    import clr

                    clr.AddReference("System")
                    clr.AddReference("System.Windows.Forms")

                    from System import IntPtr
                    from System.Windows.Forms import Screen

                    user32 = ctypes.windll.user32  # type: ignore[attr-defined]
                    rect_tuple = (0, 0, 1200, 800)
                    _maximized = False

                    def _maximize() -> None:
                        nonlocal _maximized, rect_tuple
                        if not _maximized:
                            hwnd = user32.GetForegroundWindow()
                            screen = Screen.FromHandle(IntPtr(hwnd))
                            rect = ctypes.wintypes.RECT()
                            user32.GetWindowRect(hwnd, ctypes.pointer(rect))
                            rect_tuple = (
                                rect.left,
                                rect.top,
                                rect.right - rect.left,
                                rect.bottom - rect.top,
                            )
                            user32.MoveWindow(
                                hwnd,
                                screen.Bounds.X,
                                screen.Bounds.Y,
                                screen.WorkingArea.Width,
                                screen.WorkingArea.Height,
                                True,
                            )
                            _maximized = True

                    app.native.main_window.maximize = _maximize

                    def _restore() -> None:
                        nonlocal _maximized
                        if _maximized:
                            _maximized = False
                        hwnd = user32.GetForegroundWindow()
                        user32.MoveWindow(hwnd, *rect_tuple, True)

                    app.native.main_window.restore = _restore
                maximized = False
                maximize_text = _("Maximize")
                restore_text = _("Restore")

                def toggle_maxmized() -> None:
                    nonlocal maximized
                    if maximized:
                        app.native.main_window.restore()
                        maximize_button.props("icon=open_in_full")
                        maximize_button_tooltip.set_text(maximize_text)
                    else:
                        app.native.main_window.maximize()
                        maximize_button.props("icon=close_fullscreen")
                        maximize_button_tooltip.set_text(restore_text)
                    maximized = not maximized

                ui.button(icon="minimize", on_click=app.native.main_window.minimize).classes(
                    "aspect-square"
                ).tooltip(_("Minimize"))
                with ui.button(
                    icon="open_in_full",
                    on_click=toggle_maxmized,
                ).classes("aspect-square") as maximize_button:
                    maximize_button_tooltip = ui.tooltip(maximize_text)
                ui.button(
                    icon="close",
                    color="negative",
                    on_click=app.native.main_window.destroy,
                ).classes("aspect-square").tooltip(_("Close"))
        with ui.row().classes("hidden") as submenu:
            with (
                ui.button(
                    _("Convert"),
                    icon="loop",
                    on_click=lambda: convert_menu.open(),  # type: ignore[has-type]
                )
                .props("square")
                .tooltip("Alt+C"),
                ui.menu() as convert_menu,
            ):
                with (
                    ui.menu_item(
                        on_click=selected_formats.add_upload,
                    ).tooltip("Alt+O"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("file_open").classes("text-lg")
                    ui.label(_("Import project"))
                with (
                    ui.menu_item(
                        on_click=selected_formats.batch_convert,
                    )
                    .bind_visibility_from(
                        selected_formats,
                        "task_count",
                        backward=bool,
                    )
                    .tooltip("Alt+Enter"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("play_arrow").classes("text-lg")
                    ui.label(_("Convert"))
                with (
                    ui.menu_item(on_click=selected_formats.reset)
                    .bind_visibility_from(
                        selected_formats,
                        "task_count",
                        backward=bool,
                    )
                    .tooltip("Alt+/"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("refresh").classes("text-lg")
                    ui.label(_("Clear Task List"))
                ui.separator()
                with (
                    ui.menu_item(on_click=swap_values).tooltip("Alt+\\"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("swap_vert").classes("text-lg")
                    ui.label(_("Swap Input and Output"))
            with (
                ui.button(
                    _("Import format"),
                    icon="login",
                    on_click=lambda: input_formats_menu.open(),  # type: ignore[has-type]
                )
                .props("square")
                .tooltip("Alt+["),
                ui.menu() as input_formats_menu,
            ):
                input_format_item = (
                    ui.radio(
                        {
                            k: f"{i} " + _(v["file_format"] or "") + " " + (v["suffix"] or "")
                            for i, (k, v) in enumerate(plugin_details.items())
                        },
                    )
                    .bind_value(selected_formats, "input_format")
                    .classes("text-sm")
                )
            with (
                ui.button(
                    _("Export format"),
                    icon="logout",
                    on_click=lambda: output_formats_menu.open(),  # type: ignore[has-type]
                )
                .props("square")
                .tooltip("Alt+]"),
                ui.menu() as output_formats_menu,
            ):
                output_format_item = ui.radio(
                    {
                        k: f"{i} " + _(v["file_format"] or "") + " " + (v["suffix"] or "")
                        for i, (k, v) in enumerate(plugin_details.items())
                    },
                ).bind_value(selected_formats, "output_format")
            with (
                ui.button(
                    _("Switch Theme"),
                    icon="palette",
                    on_click=lambda: theme_menu.open(),  # type: ignore[has-type]
                ).tooltip("Alt+T"),
                ui.menu() as theme_menu,
            ):
                with (
                    ui.menu_item(on_click=dark_toggler.disable).tooltip("Alt+W"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("light_mode").classes("text-lg")
                    ui.label(_("Light"))
                with (
                    ui.menu_item(on_click=dark_toggler.enable).tooltip("Alt+B"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("dark_mode").classes("text-lg")
                    ui.label(_("Dark"))
                with (
                    ui.menu_item(on_click=dark_toggler.auto).tooltip("Alt+A"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("brightness_auto").classes("text-lg")
                    ui.label(_("System"))
            ui.button(
                _("Settings"),
                icon="settings",
                on_click=settings_dialog.open,
            ).tooltip("Alt+S")
            with (
                ui.button(
                    _("Help"),
                    icon="help",
                    on_click=lambda: help_menu.open(),  # type: ignore[has-type]
                )
                .props("square")
                .tooltip("Alt+H"),
                ui.menu() as help_menu,
            ):
                with (
                    ui.menu_item(on_click=about_dialog.open).tooltip("Alt+I"),
                    ui.row().classes("items-center"),
                ):
                    ui.icon("info").classes("text-lg")
                    ui.label(_("About"))
                with ui.menu_item(), ui.row().classes("items-center"):
                    ui.icon("text_snippet").classes("text-lg")
                    ui.link(
                        _("Documentation"),
                        target="https://soulmelody.github.io/LibreSVIP",
                        new_tab=True,
                    )

        async def handle_key(e: KeyEventArguments) -> None:
            if e.modifiers.alt or e.modifiers.ctrl or e.modifiers.meta or e.modifiers.shift:
                if e.modifiers.alt and e.action.keyup and not e.action.repeat:
                    match e.key.name:
                        case "c":
                            convert_menu.open()
                        case "[":
                            input_formats_menu.open()
                        case "]":
                            output_formats_menu.open()
                        case "t":
                            theme_menu.open()
                        case "h":
                            help_menu.open()
                        case "o":
                            await selected_formats.add_upload()
                        case "i":
                            about_dialog.open()
                        case "\\":
                            swap_values()
                        case "/":
                            selected_formats.reset()
                        case "w":
                            dark_toggler.disable()
                        case "b":
                            dark_toggler.enable()
                        case "a":
                            dark_toggler.auto()
                        case "s":
                            settings_dialog.open()
                        case "Enter":
                            await selected_formats.batch_convert()
            elif e.key.number is not None and not e.action.repeat and e.action.keyup:
                key = e.key.number
                for formats_menu, format_item in [
                    (input_formats_menu, input_format_item),
                    (output_formats_menu, output_format_item),
                ]:
                    current_index = format_item._value_to_model_value(format_item.value)
                    count = len(format_item._values)
                    if formats_menu.value:
                        if count >= 10 + key:
                            next_focus = 10 * ((current_index // 10) + 1)
                            next_focus = key if next_focus + key >= count else next_focus + key
                            format_item.value = format_item._values[next_focus]
                        elif current_index != key:
                            format_item.value = format_item._values[key]

        ui.keyboard(on_key=handle_key, active=True)
    uploader = ui.upload(
        multiple=True,
        on_upload=selected_formats.add_task,
        auto_upload=True,
    ).props("hidden")

    def file_format_area() -> None:
        nonlocal select_input, select_output
        with ui.column().classes("w-full"):
            ui.label(_("Choose file format")).classes(
                "text-h5 font-bold",
            )
            with ui.grid().classes(
                "grid grid-cols-11 gap-4 w-full",
            ):
                select_input = (
                    ui.select(
                        {
                            k: _(v["file_format"] or "") + " " + (v["suffix"] or "")
                            for k, v in plugin_details.items()
                        },
                        label=_("Import format"),
                    )
                    .classes("col-span-10")
                    .bind_value(selected_formats, "input_format")
                )
                with ui.dialog() as input_info, ui.card():
                    input_plugin_info()
                    with (
                        ui.card_actions()
                        .props(
                            "align=right",
                        )
                        .classes("w-full")
                    ):
                        ui.button(
                            _("Close"),
                            on_click=input_info.close,
                        )
                ui.button(
                    icon="info",
                    on_click=input_info.open,
                ).classes(
                    "min-w-[45px] max-w-[45px] aspect-square",
                ).tooltip(_("View Detail Information"))
                ui.switch(_("Auto detect import format")).classes(
                    "col-span-5",
                ).bind_value(
                    settings,
                    "auto_detect_input_format",
                )
                ui.switch(
                    _("Reset list when import format changed"),
                ).classes("col-span-5").bind_value(
                    settings,
                    "reset_tasks_on_input_change",
                )
                ui.button(
                    icon="swap_vert",
                    on_click=swap_values,
                ).classes("min-w-[45px] max-w-[45px] w-fit aspect-square").props("round").tooltip(
                    _("Swap Input and Output")
                )
                select_output = (
                    ui.select(
                        {
                            k: _(v["file_format"] or "") + " " + (v["suffix"] or "")
                            for k, v in plugin_details.items()
                        },
                        label=_("Export format"),
                    )
                    .classes("col-span-10")
                    .bind_value(selected_formats, "output_format")
                )
                with (
                    ui.dialog().classes(
                        "h-400 w-600",
                    ) as output_info,
                    ui.card(),
                ):
                    output_plugin_info()
                    with (
                        ui.card_actions()
                        .props(
                            "align=right",
                        )
                        .classes("w-full")
                    ):
                        ui.button(
                            _("Close"),
                            on_click=output_info.close,
                        )
                ui.button(
                    icon="info",
                    on_click=output_info.open,
                ).classes(
                    "min-w-[45px] max-w-[45px] aspect-square",
                ).tooltip(_("View Detail Information"))

    def tasks_area() -> None:
        with ui.card().classes("w-full h-full").tight() as tasks_card:
            with ui.row().classes("w-full"):
                ui.label(_("Import project")).classes("text-h5 font-bold")
                ui.space()
                ui.label(_("Conversion Mode:")).classes("text-h5")
                ui.toggle(
                    {mode.value: mode.name for mode in ConversionMode},
                ).bind_value(
                    selected_formats,
                    "conversion_mode",
                )
            selected_formats.tasks_container()
            tasks_card.bind_visibility_from(
                selected_formats,
                "task_count",
                backward=bool,
            )
            tasks_card.on(
                "dragover",
                js_handler="""(event) => {
                event.preventDefault()
            }""",
            )
            tasks_card.on(
                "drop",
                js_handler=f"""(event) => {{
                for (let file of event.dataTransfer.files) {{
                    let file_name = file.name
                    post_form('{uploader.props["url"]}', {{
                        file_name: file
                    }})
                }}
                event.preventDefault()
            }}""",
            )
            with QFab(
                icon="construction",
            ).classes("absolute bottom-0 left-0 m-2 z-10") as fab:
                with fab.add_slot("active-icon"):
                    ui.icon("construction").classes("rotate-45")
                fab.on(
                    "mouseenter",
                    functools.partial(fab.run_method, "show"),
                )
                QFabAction(
                    icon="refresh",
                    on_click=selected_formats.reset,
                ).tooltip(_("Clear Task List"))
                QFabAction(
                    icon="filter_alt_off",
                    on_click=selected_formats.filter_input_ext,
                ).tooltip(_("Remove Tasks With Other Extensions"))
            with ui.row().classes(
                "absolute bottom-0 right-2 m-2 z-10",
            ):
                ui.label(_("Max Track count:")).bind_visibility_from(
                    selected_formats,
                    "conversion_mode",
                    backward=ConversionMode.SPLIT.value.__eq__,
                )
                ui.knob(
                    min=1,
                    max=10,
                    step=1,
                    show_value=True,
                    track_color="light-blue",
                ).bind_value(settings, "max_track_count").bind_visibility_from(
                    selected_formats,
                    "conversion_mode",
                    backward=ConversionMode.SPLIT.value.__eq__,
                )
                with ui.button(
                    icon="add",
                    on_click=selected_formats.add_upload,
                ).props("round"):
                    ui.badge().props(
                        "floating color=orange",
                    ).bind_text_from(
                        selected_formats,
                        "task_count",
                        backward=str,
                    ).tooltip(_("Continue Adding files"))
        with (
            ui.card()
            .classes(
                "w-full h-full opacity-60 hover:opacity-100 flex items-center justify-center border-dashed border-2 border-indigo-300 hover:border-indigo-500",
            )
            .style("cursor: pointer")
            .tight() as upload_card
        ):
            upload_card.on(
                "dragover",
                js_handler="""(event) => {
                event.preventDefault()
            }""",
            )
            upload_card.on(
                "drop",
                js_handler=f"""(event) => {{
                for (let file of event.dataTransfer.files) {{
                    let file_name = file.name
                    post_form('{uploader.props["url"]}', {{
                        file_name: file
                    }})
                }}
                event.preventDefault()
            }}""",
            )
            upload_card.on("click", selected_formats.add_upload)
            upload_card.bind_visibility_from(
                selected_formats,
                "task_count",
                backward=not_,
            )
            ui.icon("file_upload").classes("text-6xl")
            ui.label(
                _("Drag and drop files here or click to upload"),
            ).classes("text-lg")

    @ui.refreshable
    def middleware_options() -> None:
        for middleware in middleware_manager.plugin_registry.values():
            with ui.row().classes("items-center w-full"):
                middleware_toggler = (
                    ui.switch(_(middleware.name))
                    .props("color=green")
                    .bind_value(
                        selected_formats.middleware_enabled_states,
                        middleware.identifier,
                    )
                )
                if middleware.description:
                    ui.space()
                    ui.icon("help_outline").classes("text-3xl").style(
                        "cursor: help",
                    ).tooltip(_(middleware.description))
            middleware_options_form(middleware.identifier, middleware_toggler)

    def options_area() -> None:
        with ui.scroll_area().classes("w-full h-full"):
            with ui.row().classes("absolute top-0 right-2 m-2 z-10"):
                ui.button(
                    icon="play_arrow",
                    on_click=selected_formats.batch_convert,
                ).props("round").bind_visibility_from(
                    selected_formats,
                    "task_count",
                    backward=bool,
                ).tooltip(_("Start Conversion"))
                ui.button(
                    icon="download_for_offline",
                    on_click=selected_formats.save_file,
                ).props("round").bind_visibility_from(
                    selected_formats,
                    "task_count",
                    backward=bool,
                ).tooltip(_("Export"))
            ui.label(_("Advanced Options")).classes("text-h5 font-bold")
            with ui.expansion().classes("w-full") as import_panel:
                with import_panel.add_slot("header"):
                    input_panel_header()
                input_options()
            ui.separator()
            with ui.expansion().classes("w-full") as middleware_panel:
                with (
                    middleware_panel.add_slot("header"),
                    ui.row().classes("w-full items-center"),
                ):
                    ui.icon("auto_fix_high").classes("text-lg")
                    ui.label(_("Intermediate Processing")).classes("text-subtitle1 font-bold")
                middleware_options()
            ui.separator()
            with ui.expansion().classes("w-full") as export_panel:
                with export_panel.add_slot("header"):
                    output_panel_header()
                output_options()

    with ui.card().classes("w-full min-w-80").style("height: calc(100vh - 155px)").tight():
        with ui.splitter(limits=(40, 60)).classes(
            "w-full h-0 sm:invisible lg:h-full lg:visible"
        ) as main_splitter:
            with (
                main_splitter.before,
                ui.splitter(limits=(40, 50), horizontal=True) as left_splitter,
            ):
                with left_splitter.before, ui.card().classes("h-full w-full"):
                    file_format_area()
                with left_splitter.after:
                    tasks_area()
            with main_splitter.after:
                options_area()
        with (
            ui.card().classes("w-full h-full sm:visible lg:h-0 lg:invisible").tight(),
            ui.tab_panels(tabs, value=format_select_tab).classes("h-full w-full"),
        ):
            with (
                ui.tab_panel(format_select_tab),
                ui.splitter(limits=(60, 60), horizontal=True, value=60).classes(
                    "h-full w-full"
                ) as format_select_splitter,
            ):
                with (
                    format_select_splitter.before,
                    ui.card().classes("h-full w-full").tight(),
                ):
                    file_format_area()
                with format_select_splitter.after:
                    tasks_area()
            with (
                ui.tab_panel(options_tab),
                ui.splitter(limits=(50, 60), horizontal=True).classes(
                    "h-full w-full"
                ) as options_splitter,
            ):
                with options_splitter.before:
                    options_area()
                with options_splitter.after:
                    tasks_area()
    with ui.footer().classes("bg-transparent"):
        ajax_bar = ui.element("q-ajax-bar").props("position=bottom")

    def add_javascript() -> None:
        nonlocal select_input
        ui.add_body_html(
            textwrap.dedent(
                f"""
            <script>
                function get_element(element_id) {{
                    let element = getElement(element_id)
                    if (element.$refs.qRef !== undefined)
                        return element.$refs.qRef
                    return element
                }}
                function post_form(url, data) {{
                    let ajax_bar = get_element({ajax_bar.id})
                    let form_data = new FormData()
                    for (let key in data) {{
                        form_data.append(key, data[key])
                    }}
                    var xhr = new XMLHttpRequest();
                    var progress = 0
                    xhr.onreadystatechange = function() {{
                        if (xhr.readyState === 4) {{
                            ajax_bar.stop()
                        }}
                    }}
                    xhr.upload.onprogress = function(event) {{
                        if (event.lengthComputable) {{
                            let next = event.loaded / event.total * 100
                            if (next - progress > 0.5) {{
                                ajax_bar.increment(next - progress)
                                progress = next
                            }}
                        }}
                    }}
                    xhr.open('POST', url, true);
                    ajax_bar.start()
                    xhr.send(form_data);
                }}
                function add_upload() {{
                    if (window.showOpenFilePicker) {{
                        let format_desc = get_element({select_input.id}).modelValue.label
                        let suffix = format_desc.match(/\\((?:\\*)(\\..*?)\\)/)[1]
                        let bracket_index = format_desc.lastIndexOf('(')
                        let file_format = format_desc.substr(0, bracket_index === -1 ? format_desc.length : bracket_index)
                        window.showOpenFilePicker(
                            {{
                                types: [
                                    {{
                                        description: file_format,
                                        accept: {{
                                            '*/*': [suffix],
                                        }}
                                    }}
                                ],
                                multiple: true
                            }}
                        ).then(async function (fileHandles) {{
                            for (const fileHandle of fileHandles) {{
                                const file = await fileHandle.getFile();
                                let file_name = file.name
                                post_form('{uploader.props["url"]}', {{
                                    file_name: file
                                }})
                            }}
                        }});
                    }} else {{
                        get_element({uploader.id}).pickFiles()
                    }}
                }}
            </script>
            """,
            ).strip(),
        )

    add_javascript()


def main() -> None:
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--host", type=str, default="127.0.0.1")
    arg_parser.add_argument("--port", type=int, default=8080)
    arg_parser.add_argument("--server", action="store_true")
    arg_parser.add_argument("--daemon", action="store_true")
    args, argv = arg_parser.parse_known_args()

    if shutil.which("termux-open-url") is not None:
        # a workaround for termux platform, from https://github.com/python/cpython/issues/90371#issuecomment-1460738762
        webbrowser.register("termux-open-url '%s'", None)

    secrets_path = app_dir.user_config_path / "secrets.txt"
    if not secrets_path.exists():
        secrets_path.parent.mkdir(parents=True, exist_ok=True)
        secrets_path.write_text(secrets.token_urlsafe(32))
    storage_secret = secrets_path.read_text()

    with as_file(res_dir / "libresvip.ico") as icon_path:
        ui.run(
            show=not args.daemon,
            window_size=None if args.server else (1200, 800),
            frameless=not args.server,
            reload=False,
            host=args.host if args.server else None,
            port=args.port,
            storage_secret=storage_secret,
            title="LibreSVIP",
            favicon=icon_path,
        )



from __future__ import annotations

import sys
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .model import USTXProject


@dataclass
class TimeSigSegment:
    bar_pos: int
    tick_pos: int
    beat_per_bar: int
    beat_unit: int
    ticks_per_bar: int
    ticks_per_beat: int
    bar_end: int = sys.maxsize
    tick_end: int = sys.maxsize


@dataclass
class TempoSegment:
    tick_pos: int
    beat_per_bar: int
    beat_unit: int
    bpm: float = 120
    ms_pos: float = 0
    ms_per_tick: float = 0
    ticks_per_ms: float = 0
    tick_end: int = sys.maxsize
    ms_end: float = float("inf")

    @property
    def ticks(self) -> int:
        return self.tick_end - self.tick_pos


@dataclass
class TimeAxis:
    time_sig_segments: list[TimeSigSegment] = field(default_factory=list)
    tempo_segments: list[TempoSegment] = field(default_factory=list)

    def build_segments(self, project: USTXProject) -> None:
        self.time_sig_segments.clear()
        for i in range(len(project.time_signatures)):
            timesig = project.time_signatures[i]
            pos_tick = 0
            if i > 0:
                last_bar_pos = project.time_signatures[i - 1].bar_position
                pos_tick = self.time_sig_segments[-1].tick_pos + self.time_sig_segments[
                    -1
                ].ticks_per_bar * (timesig.bar_position - last_bar_pos)
            else:
                assert timesig.bar_position == 0
            self.time_sig_segments.append(
                TimeSigSegment(
                    bar_pos=timesig.bar_position or 0,
                    tick_pos=pos_tick,
                    beat_per_bar=timesig.beat_per_bar,
                    beat_unit=timesig.beat_unit,
                    ticks_per_bar=project.resolution
                    * 4
                    * timesig.beat_per_bar
                    // timesig.beat_unit,
                    ticks_per_beat=project.resolution * 4 // timesig.beat_unit,
                )
            )
        for i in range(len(self.time_sig_segments) - 1):
            self.time_sig_segments[i].bar_end = self.time_sig_segments[i + 1].bar_pos
            self.time_sig_segments[i].tick_end = self.time_sig_segments[i + 1].tick_pos
        if not self.time_sig_segments:
            self.time_sig_segments.append(
                TimeSigSegment(
                    bar_pos=0,
                    tick_pos=0,
                    beat_per_bar=project.beat_per_bar,
                    beat_unit=project.beat_unit,
                    ticks_per_bar=project.resolution
                    * 4
                    * project.beat_per_bar
                    // project.beat_unit,
                    ticks_per_beat=project.resolution * 4 // project.beat_unit,
                )
            )

        self.tempo_segments.clear()
        self.tempo_segments.extend(
            [
                TempoSegment(
                    tick_pos=sigseg.tick_pos,
                    beat_per_bar=sigseg.beat_per_bar,
                    beat_unit=sigseg.beat_unit,
                )
                for sigseg in self.time_sig_segments
            ]
        )
        for i in range(len(project.tempos)):
            tempo = project.tempos[i]
            if i == 0:
                assert tempo.position == 0
            index = next(
                (j for j, seg in enumerate(self.tempo_segments) if seg.tick_pos >= tempo.position),
                -1,
            )
            if index < 0:
                self.tempo_segments.append(
                    TempoSegment(
                        tick_pos=tempo.position,
                        bpm=tempo.bpm,
                        beat_per_bar=self.tempo_segments[-1].beat_per_bar,
                        beat_unit=self.tempo_segments[-1].beat_unit,
                    )
                )
            elif self.tempo_segments[index].tick_pos == tempo.position:
                self.tempo_segments[index].bpm = tempo.bpm
            else:
                self.tempo_segments.insert(
                    index,
                    TempoSegment(
                        tick_pos=tempo.position,
                        bpm=tempo.bpm,
                        beat_per_bar=self.tempo_segments[index - 1].beat_per_bar,
                        beat_unit=self.tempo_segments[index - 1].beat_unit,
                    ),
                )
        for i in range(len(self.tempo_segments) - 1):
            if self.tempo_segments[i + 1].bpm == 0:
                self.tempo_segments[i + 1].bpm = self.tempo_segments[i].bpm
            self.tempo_segments[i].tick_end = self.tempo_segments[i + 1].tick_pos
        for i in range(len(self.tempo_segments)):
            self.tempo_segments[i].ms_per_tick = (
                60.0
                * 1000.0
                * self.tempo_segments[i].beat_per_bar
                / (self.tempo_segments[i].bpm * 4 * project.resolution)
            )
            self.tempo_segments[i].ticks_per_ms = (
                self.tempo_segments[i].bpm
                * 4
                * project.resolution
                / (60.0 * 1000.0 * self.tempo_segments[i].beat_per_bar)
            )
            if i > 0:
                self.tempo_segments[i].ms_pos = (
                    self.tempo_segments[i - 1].ms_pos
                    + self.tempo_segments[i - 1].ticks * self.tempo_segments[i - 1].ms_per_tick
                )
                self.tempo_segments[i - 1].ms_end = self.tempo_segments[i].ms_pos

    def tick_pos_to_ms_pos(self, tick: float) -> float:
        segment = next(
            (seg for seg in self.tempo_segments if seg.tick_pos == tick or seg.tick_end > tick),
            None,
        )
        return segment.ms_pos + segment.ms_per_tick * (tick - segment.tick_pos) if segment else 0

    def ms_pos_to_tick_pos(self, ms: float) -> int:
        segment = next(
            (seg for seg in self.tempo_segments if seg.ms_pos == ms or seg.ms_end > ms),
            None,
        )
        tick_pos = segment.tick_pos + (ms - segment.ms_pos) * segment.ticks_per_ms if segment else 0
        return round(tick_pos)

    def ms_between_tick_pos(self, tick_pos: float, tick_end: float) -> float:
        return self.tick_pos_to_ms_pos(tick_end) - self.tick_pos_to_ms_pos(tick_pos)



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
    StaticTempoMixin,
)
from libresvip.utils.translation import gettext_lazy as _

from .enums import PocketSingerLyricsLanguage


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    ExtractEmbededAudioMixin,
    BaseModel,
):
    pass


class OutputOptions(StaticTempoMixin, BaseModel):
    lyric_language: PocketSingerLyricsLanguage = Field(
        default=PocketSingerLyricsLanguage.CHINESE,
        title=_("Lyrics language"),
    )
    title: str = Field(
        default="New Project",
        title=_("Title of the project"),
    )



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.model.reset_time_axis import reset_time_axis

from .diffsinger_generator import DiffSingerGenerator
from .diffsinger_parser import DiffSingerParser
from .options import InputOptions, OutputOptions
from .utils.models.ds_file import DsProject
from .utils.project_util import split_into_segments


class DiffSingerConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        ds_project = DsProject.model_validate_json(path.read_bytes().decode("utf-8"))
        return DiffSingerParser(options).parse_project(ds_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        if options.split_threshold >= 0:
            segments = split_into_segments(
                project,
                options.min_interval,
                int(options.split_threshold * 1000),
            )
            series = []
            for segment in segments:
                ds_params = DiffSingerGenerator(
                    options=options, trailing_space=segment[2]
                ).generate(segment[1])
                ds_params.offset = segment[0]
                if options.seed >= 0:
                    ds_params.seed = options.seed
                series.append(ds_params)
            ds_project = DsProject(root=series)
        else:
            project = reset_time_axis(project)
            diff_singer_params = DiffSingerGenerator(options=options, trailing_space=0.5).generate(
                project
            )
            if options.seed >= 0:
                diff_singer_params.seed = options.seed
            ds_project = DsProject(root=[diff_singer_params])
        path.write_bytes(
            json.dumps(
                ds_project.model_dump(mode="json"),
                indent=options.indent,
                ensure_ascii=False,
            ).encode("utf-8")
        )



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import StaticTempoMixin
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(BaseModel):
    pass


class OutputOptions(StaticTempoMixin, BaseModel):
    singer_name: str = Field(
        title=_("Singer name"),
        description=_("Please enter the singer's English name."),
        default="Doaz",
    )



import re
from collections.abc import Iterable

import pypinyin
import zhon

from libresvip.utils.text import LATIN_ALPHABET

CHINESE_RE: re.Pattern[str] = re.compile(rf"[{zhon.hanzi.characters}]")
WHITE_SPACE: re.Pattern[str] = re.compile(r"[\s\n\r\t]+")


def get_pinyin_series(
    chinese_series: Iterable[str],
    ignore_hyphens: bool = True,
    reverse_letters: bool = True,
    filter_non_chinese: bool = True,
) -> list[str]:
    chinese_list = list(chinese_series)
    pinyin_list = ["" for _ in range(len(chinese_list))]
    item_counts = [0 for _ in range(len(chinese_list))]
    result_items = []
    chinese = ""
    for i, part in enumerate(chinese_list):
        count = 0
        is_letter = False
        non_chinese = ""
        for char in part:
            if WHITE_SPACE.match(char) is not None or (ignore_hyphens and char == "-"):
                if non_chinese:
                    result_items.append(non_chinese)
                    count += 1
                    non_chinese = ""
            elif CHINESE_RE.match(char) is not None:
                if len(non_chinese):
                    result_items.append(non_chinese)
                    count += 1
                    non_chinese = ""
                chinese += char
                count += 1
            else:
                if len(chinese):
                    result_items.extend(pypinyin.lazy_pinyin(chinese, errors=lambda x: " "))
                    chinese = ""
                if LATIN_ALPHABET.match(char) is not None:
                    if reverse_letters:
                        if len(non_chinese) and not is_letter:
                            result_items.append(non_chinese)
                            count += 1
                            non_chinese = ""
                        non_chinese += char
                        is_letter = True
                elif not filter_non_chinese:
                    if len(non_chinese) and is_letter:
                        result_items.append(non_chinese)
                        count += 1
                        non_chinese = ""
                    non_chinese += char
                    is_letter = False
        if len(non_chinese):
            result_items.append(non_chinese)
            count += 1
        item_counts[i] = count
    if len(chinese):
        result_items.extend(pypinyin.lazy_pinyin(chinese, errors=lambda x: " "))
    pinyin = ""
    index = 0
    for i in range(len(pinyin_list)):
        if item_counts[i] == 0:
            pinyin_list[i] = ""
            continue
        pinyin += result_items[index]
        index += 1
        for _ in range(1, item_counts[i]):
            if result_items[index - 1][-1].isalnum() and result_items[index][0].isalnum():
                pinyin += " "
            pinyin += result_items[index]
            index += 1
        pinyin_list[i] = pinyin
        pinyin = ""
    return pinyin_list



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .aisingers_generator import AiSingersGenerator
from .aisingers_parser import AiSingersParser
from .model import AISProjectBody, AISProjectHead
from .options import InputOptions, OutputOptions


class AiSingersConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        content = path.read_bytes().decode("utf-8")
        head, _, body = content.partition("\n")
        ais_project_head = AISProjectHead.model_validate_json(head.strip())
        ais_project_body = AISProjectBody.model_validate_json(body.strip())
        return AiSingersParser(options).parse_project(ais_project_head, ais_project_body)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ais_project_head, ais_project_body = AiSingersGenerator(options).generate_project(project)
        path.write_bytes(
            (
                json.dumps(
                    ais_project_head.model_dump(mode="json", by_alias=True),
                    separators=(", ", ": "),
                )
                + "\n"
                + json.dumps(
                    ais_project_body.model_dump(mode="json", by_alias=True),
                    separators=(", ", ": "),
                )
            ).encode("utf-8")
        )



import dataclasses
import pathlib

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.lyric_phoneme.chinese.vocaloid_xsampa import pinyin2xsampa
from libresvip.core.lyric_phoneme.japanese import to_romaji
from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import (
    legato_chars,
    romaji2xsampa,
)
from libresvip.core.lyric_phoneme.korean.vocaloid_xsampa import hangul2xsampa
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.translation import gettext_lazy as _

from .constants import (
    BPM_RATE,
    DEFAULT_CHINESE_PHONEME,
    DEFAULT_JAPANESE_PHONEME,
    PITCH_BEND_NAME,
    PITCH_BEND_SENSITIVITY_NAME,
)
from .model import (
    VocaloidAITrack,
    VocaloidAIVoice,
    VocaloidAudioTrack,
    VocaloidControllers,
    VocaloidLangID,
    VocaloidLanguage,
    VocaloidNotes,
    VocaloidPoint,
    VocaloidProject,
    VocaloidRegion,
    VocaloidStandardTrack,
    VocaloidTimeSig,
    VocaloidTracks,
    VocaloidVoice,
    VocaloidVoicePart,
    VocaloidVoices,
    VocaloidWav,
    VocaloidWavPart,
)
from .options import OutputOptions
from .vocaloid_pitch import generate_for_vocaloid


@dataclasses.dataclass
class VocaloidGenerator:
    options: OutputOptions
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    end_tick: int = 0
    wav_paths: dict[str, pathlib.Path] = dataclasses.field(default_factory=dict)

    def generate_project(self, project: Project) -> VocaloidProject:
        vpr = VocaloidProject(
            voices=[
                VocaloidVoices(
                    comp_id=self.options.default_comp_id,
                    name=self.options.default_singer_name,
                )
            ]
        )
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        vpr.master_track.time_sig.events = self.generate_time_signatures(
            project.time_signature_list
        )
        vpr.master_track.tempo.events = self.generate_tempos(project.song_tempo_list)
        vpr.master_track.volume.events.append(VocaloidPoint(pos=0, value=0))
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        vpr.tracks = self.generate_tracks(project.track_list)
        vpr.master_track.loop.end = self.end_tick
        return vpr

    def generate_time_signatures(
        self, time_signature_list: list[TimeSignature]
    ) -> list[VocaloidTimeSig]:
        output_tick = 0.0
        time_sig_events: list[VocaloidTimeSig] = []
        for time_signature in time_signature_list:
            if not time_sig_events:
                output_tick += TICKS_IN_BEAT * 4 * time_signature.bar_index
            else:
                output_tick += (
                    TICKS_IN_BEAT
                    * 4
                    * (time_sig_events[-1].numer / time_sig_events[-1].denom)
                    * (time_signature.bar_index - time_sig_events[-1].bar)
                )
            time_sig_events.append(
                VocaloidTimeSig(
                    bar=time_signature.bar_index,
                    denom=time_signature.denominator,
                    numer=time_signature.numerator,
                )
            )
        self.end_tick = max(self.end_tick, round(output_tick))
        return time_sig_events

    def generate_tempos(self, tempo_list: list[SongTempo]) -> list[VocaloidPoint]:
        tempo_events = [
            VocaloidPoint(pos=it.position, value=int(it.bpm * BPM_RATE)) for it in tempo_list
        ]
        self.end_tick = max(self.end_tick, max((it.pos for it in tempo_events), default=0))
        return tempo_events

    def generate_phoneme(self, lyric: str) -> str:
        if lyric in legato_chars:
            return "-"
        if self.options.default_lang_id == VocaloidLanguage.SIMPLIFIED_CHINESE:
            pinyin = " ".join(get_pinyin_series([lyric], filter_non_chinese=False))
            return pinyin2xsampa.get(pinyin, DEFAULT_CHINESE_PHONEME)
        elif self.options.default_lang_id == VocaloidLanguage.JAPANESE:
            return romaji2xsampa.get(to_romaji(lyric), DEFAULT_JAPANESE_PHONEME)
        elif self.options.default_lang_id == VocaloidLanguage.KOREAN:
            return hangul2xsampa(lyric)
        return DEFAULT_CHINESE_PHONEME

    def generate_tracks(self, track_list: list[Track]) -> list[VocaloidTracks]:
        tracks: list[VocaloidTracks] = []
        singing_track_found = False
        for track in track_list:
            if isinstance(track, InstrumentalTrack):
                wav_path = pathlib.Path(track.audio_file_path)
                if (
                    (track_info := audio_track_info(track.audio_file_path, only_wav=True))
                    is not None
                    and track_info.sampling_rate == 44100
                    and track_info.bit_depth == 16
                ):
                    audio_duration_in_secs = track_info.duration / 1000
                    wav_part_region_end = self.time_synchronizer.get_actual_ticks_from_secs_offset(
                        track.offset, audio_duration_in_secs
                    )
                    self.wav_paths[wav_path.name] = wav_path
                    wav_part = VocaloidWavPart(
                        pos=track.offset,
                        wav=VocaloidWav(
                            original_name=wav_path.name,
                            name=wav_path.name,
                        ),
                        region=VocaloidRegion(
                            begin=0,
                            end=wav_part_region_end,
                        ),
                    )
                    self.end_tick = max(
                        self.end_tick,
                        int(wav_part_region_end),
                    )
                    tracks.append(
                        VocaloidAudioTrack(
                            name=track.title,
                            parts=[wav_part],
                            is_muted=track.mute,
                            is_solo_mode=track.solo,
                        )
                    )
            else:
                singing_track_found = True
                notes = [
                    VocaloidNotes(
                        pos=note.start_pos,
                        duration=note.length,
                        number=note.key_number,
                        lyric=note.lyric,
                        phoneme=self.generate_phoneme(note.lyric),
                        lang_id=self.options.default_lang_id,
                    )
                    for note in track.note_list
                ]
                duration = track.note_list[-1].end_pos if track.note_list else None
                controllers = self.generate_pitch_data(track)
                part = (
                    VocaloidVoicePart(
                        duration=duration,
                        notes=notes,
                        controllers=controllers,
                    )
                    if duration
                    else None
                )
                if part is not None:
                    if self.options.is_ai_singer:
                        part.ai_voice = VocaloidAIVoice(
                            comp_id=self.options.default_comp_id,
                            lang_ids=[VocaloidLangID(lang_id=self.options.default_lang_id)],
                        )
                    else:
                        part.voice = VocaloidVoice(
                            comp_id=self.options.default_comp_id,
                            lang_id=self.options.default_lang_id,
                        )
                vocaloid_track = (
                    VocaloidAITrack if self.options.is_ai_singer else VocaloidStandardTrack
                )(
                    name=track.title,
                    parts=[part] if part else [],
                    is_muted=track.mute,
                    is_solo_mode=track.solo,
                )
                vocaloid_track.panpot.events.append(VocaloidPoint(pos=0, value=0))
                tracks.append(vocaloid_track)
                if duration:
                    self.end_tick = max(
                        self.end_tick,
                        duration,
                    )
        if singing_track_found and self.options.default_lang_id not in [
            VocaloidLanguage.SIMPLIFIED_CHINESE,
            VocaloidLanguage.JAPANESE,
            VocaloidLanguage.KOREAN,
        ]:
            show_warning(
                _(
                    'Phonemes of all notes were set to "la". Please use "Job" -> "Convert Phonemes to Match Languages in the menu of VOCALOID to reset them.'
                )
            )
        return tracks

    def generate_pitch_data(self, track: SingingTrack) -> list[VocaloidControllers] | None:
        raw_pitch_data = generate_for_vocaloid(
            track.edited_params.pitch,
            track.note_list,
            self.first_bar_length,
            self.time_synchronizer,
        )
        if not raw_pitch_data:
            return None
        controllers = []
        if raw_pitch_data.pbs:
            controller_events = [
                VocaloidPoint(pos=pbs_event.pos, value=pbs_event.value)
                for pbs_event in raw_pitch_data.pbs
            ]
            controllers.append(
                VocaloidControllers(name=PITCH_BEND_SENSITIVITY_NAME, events=controller_events)
            )
        if raw_pitch_data.pit:
            controller_events = [
                VocaloidPoint(pos=pit_event.pos, value=pit_event.value)
                for pit_event in raw_pitch_data.pit
            ]
            controllers.append(VocaloidControllers(name=PITCH_BEND_NAME, events=controller_events))
        return controllers
